# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import os
import shutil
import time
from typing import Literal, Optional, Union

import dask.bag as db
import dask.dataframe as dd

from nemo_curator.datasets import DocumentDataset
from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix
from nemo_curator.utils.file_utils import expand_outdir_and_mkdir
from nemo_curator.utils.semdedup_utils import (
    get_semantic_matches_per_cluster,
    prune_single_cluster,
    write_pruned_summary_file,
)


class SemanticClusterLevelDedup:
    def __init__(
        self,
        n_clusters: int = 1000,
        emb_by_clust_dir: str = "./clustering_results/embs_by_nearest_center",
        id_column: str = "id",
        which_to_keep: str = "hard",
        sim_metric: Literal["cosine", "l2"] = "cosine",
        output_dir: str = "./clustering_results",
        embedding_column: str = "embeddings",
        batched_cosine_similarity: int = 1024,
        logger: Union[logging.Logger, str] = "./",
        profile_dir: Optional[str] = None,
    ) -> None:
        """
        Initialize the SemanticClusterLevelDedup class.

        Args:
            n_clusters (int): Number of clusters. Default is 1000.
            emb_by_clust_dir (str): Directory containing embeddings by cluster.
                Default is "./clustering_results/embs_by_nearest_center".
            id_column (str): Column name used as the identifier in the dataset.
                Default is "id".
            which_to_keep (str): Method to determine which duplicates to keep. Default is "hard".
                - hard retains edge-case or outlier items farthest from the centroid by sorting points by decreasing distance from the centroid.
                - easy retains representative items closest to the centroid by sorting points by increasing distance from the centroid.
                - random retains items randomly.
            sim_metric ("cosine" or "l2"): Similarity metric to use to rank within cluster. Default is "cosine".
                `which_to_keep` determines how points within each cluster are ranked, based on the similarity to the centroid defined by `sim_metric`
            output_dir (str): Directory to save output files.
                Default is "./clustering_results".
            embedding_column (str): The column name that stores the embeddings.
                Default is "embeddings".
            batched_cosine_similarity (int): Whether to use batched cosine similarity (has less memory usage).
                Default is 1024. When greater than 0, batching is used and memory requirements are O(N*B) where N is the number of items in the cluster and B is the batch size.
                When less than or equal to 0, no batching is used and memory requirements are O(N^2) where N is the number of items in the cluster.
            logger (Union[logging.Logger, str]): Existing logger to log to, or a path to a log directory.
                Default is "./".
            profile_dir (Optional[str]): If specified, directory to write Dask profile.
                Default is None.

        """
        self.n_clusters = n_clusters
        self.emb_by_clust_dir = emb_by_clust_dir
        self.id_col = id_column
        self.which_to_keep = which_to_keep
        self.sim_metric = sim_metric
        self.output_dir = output_dir
        self.semdedup_pruning_tables_dir = os.path.join(
            output_dir, "semdedup_pruning_tables"
        )
        self.computed_semantic_match_dfs = False
        self.embedding_column = embedding_column
        self.batched_cosine_similarity = batched_cosine_similarity
        self.logger = self._setup_logger(logger)
        self.profile_dir = profile_dir

    def _setup_logger(self, logger: Union[logging.Logger, str]) -> logging.Logger:
        """
        Set up the logger.

        Args:
            logger (Union[logging.Logger, str]): Logger instance or path to the log file directory.

        Returns:
            logging.Logger: Configured logger.
        """
        if isinstance(logger, str):
            return create_logger(
                rank=0,
                name="SemanticClusterLevelDedup",
                log_file=os.path.join(logger, "SemanticClusterLevelDedup.log"),
                log_level=logging.INFO,
                stdout=True,
            )
        else:
            return logger

    def compute_semantic_match_dfs(self) -> None:
        if os.path.exists(self.semdedup_pruning_tables_dir):
            self.logger.info(
                f"Removing existing directory {self.semdedup_pruning_tables_dir}"
            )
            shutil.rmtree(self.semdedup_pruning_tables_dir)
        expand_outdir_and_mkdir(self.semdedup_pruning_tables_dir)
        t0 = time.time()

        with performance_report_if_with_ts_suffix(
            self.profile_dir, "semantic-match-compute"
        ):
            tasks = db.from_sequence(
                list(range(self.n_clusters)), npartitions=self.n_clusters
            ).map(
                lambda cluster_id: get_semantic_matches_per_cluster(
                    cluster_id=cluster_id,
                    emb_by_clust_dir=self.emb_by_clust_dir,
                    id_col=self.id_col,
                    output_dir=self.semdedup_pruning_tables_dir,
                    embedding_col=self.embedding_column,
                    which_to_keep=self.which_to_keep,
                    sim_metric=self.sim_metric,
                    batched_cosine_similarity=self.batched_cosine_similarity,
                )
            )
            tasks.compute()
        self.logger.info(
            f"Time taken for Computing Semantic Matches : {time.time() - t0}"
        )
        self.computed_semantic_match_dfs = True

    def extract_dedup_data(self, eps_to_extract: float) -> DocumentDataset:
        """
        Extract similar records that are within epsilon threshold. These records can be removed from the dataset.
        Args:
            eps_to_extract (float): Epsilon threshold for extracting deduplicated data.
        Returns:
            DocumentDataset: Dataset containing list of ids that are can be removed.
        """
        if not self.computed_semantic_match_dfs:
            raise ValueError(
                "Run compute_semantic_match_dfs before calling extract_dedup_data"
            )
        assert isinstance(eps_to_extract, float), "eps_to_extract must be a float"
        output_parquet_path = os.path.join(
            self.output_dir, f"unique_ids_{eps_to_extract}.parquet"
        )

        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir,
            "extracting-pruned-from-clusters",
        ):
            results_df = dd.from_map(
                prune_single_cluster,
                range(self.n_clusters),
                id_col=self.id_col,
                emb_by_clust_dir=self.emb_by_clust_dir,
                semdedup_pruning_tables_dir=self.semdedup_pruning_tables_dir,
                eps=eps_to_extract,
            )
            results_df.to_parquet(output_parquet_path, index=False, ignore_index=True)
            self.logger.info(
                f"Time taken for Extracting Pruned Data : {time.time() - t0} and output written at {output_parquet_path}"
            )

        # Write out summary file
        output_summary_file = os.path.join(
            self.output_dir, f"dedup_summary_{eps_to_extract}.csv"
        )
        write_pruned_summary_file(
            eps=eps_to_extract,
            emb_by_clust_dir=self.emb_by_clust_dir,
            filtered_unique_ids_path=output_parquet_path,
            output_summary_file=output_summary_file,
            logger=self.logger,
        )
        return DocumentDataset.read_parquet(
            output_parquet_path, blocksize="1gb", backend="cudf"
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import List

from huggingface_hub import hf_hub_download

from nemo_curator.utils.import_utils import safe_import

COMET_IMPORT_MSG = (
    "To run QE filtering with COMET, you need to install from PyPI with: `pip install unbabel-comet`. "
    + "More information at https://github.com/Unbabel/COMET."
)
PYMARIAN_IMPORT_MSG = (
    "To run QE filtering with Cometoid/PyMarian, you need to install PyMarian. "
    + "More information at https://github.com/marian-nmt/wmt23-metrics?tab=readme-ov-file#setup."
)
comet = safe_import("comet", msg=COMET_IMPORT_MSG)
pymarian = safe_import("pymarian", msg=PYMARIAN_IMPORT_MSG)


class QEModel(ABC):
    """Abstract model for all quality estimation models for bitext."""

    def __init__(self, name: str, model, gpu=False):
        """Args:
        name (str): A string named of the model. Not directly tied to `MODEL_NAME_TO_HF_PATH` as defined in some subclasses but it is suggested.
        model: A loaded model object. The type of the object depends on the loaded model type.
        gpu (bool, optional): Whether inference is on GPU. Defaults to False.
        """
        self._name = name
        self._model = model
        self._gpu = gpu

    @classmethod
    @abstractmethod
    def load_model(cls, model_name: str):
        """An abstract method that loads the model according to a model name.

        Args:
            model_name (str): The name of the model to be loaded.
                Could be a huggingface model name, a path, or something else, depending on the implementation.
        """
        pass

    @staticmethod
    @abstractmethod
    def wrap_qe_input(src: str, tgt: str, reverse=False, **kwargs):
        """An abstract method that implements the following: given the individual source and target string of the bitext,
        wrap them into proper format that can be accepted by the underlying model.

        Args:
            src (str): Source side string of the bitext.
            tgt (str): Target side string of the bitext.
            reverse (bool, optional): Whether to reverse the source and target side of the bitext. Defaults to False.
        """
        pass

    @abstractmethod
    def predict(self, **kwargs) -> List[float]:
        """An abstract method that calls the underlying model to produce estimated quality scores.

        Returns:
            List[float]: List of quality scores.
        """
        pass


class COMETQEModel(QEModel):
    """Wrapper class for any COMET quality estimation models (https://github.com/Unbabel/COMET)."""

    MODEL_NAME_TO_HF_PATH = {
        "comet-qe": "Unbabel/wmt20-comet-qe-da",
    }

    @classmethod
    def load_model(cls, model_name: str, gpu: bool = False):
        """See parent class docstring for details on functionality and arguments."""
        path = comet.download_model(cls.MODEL_NAME_TO_HF_PATH[model_name])
        return cls(model_name, comet.load_from_checkpoint(path), gpu)

    @staticmethod
    def wrap_qe_input(src: str, tgt: str, reverse=False, **kwargs):
        """See parent class docstring for details on functionality and arguments."""
        return {"src": src, "mt": tgt} if not reverse else {"src": tgt, "mt": src}

    def predict(self, input: List, **kwargs) -> List[float]:
        """Implements quality estimation score prediction for COMET model.

        Args:
            input (List): A list of bitext pairs wrapped as dictionaries.

        Returns:
            List[float]: List of quality scores.
        """
        return self._model.predict(
            input, gpus=int(self._gpu), num_workers=0
        ).scores  # it's critical to set num_workers=0 to avoid spawning new processes within a dask worker


class PyMarianQEModel(QEModel):

    MODEL_NAME_TO_HF_PATH = {
        "cometoid-wmt23": "marian-nmt/cometoid22-wmt23",
        "cometoid-wmt23-mqm": "marian-nmt/cometoid22-wmt23",
    }
    # Because PyMarian depends on its own deep learning library rather than PyTorch/Huggingface
    # there is unfortunately no model configuration interface that can automatically adapt to
    # individual systems (like hf `AutoConfig`).
    # Those should work on most systems, but if not please adjust as needed.
    MARIAN_GPU_ARGS = " -w 8000 --mini-batch 32 -d 0"
    MARIAN_CPU_ARGS = " --cpu-threads 1 -w 2000"
    # PyMarian has memory leakage when a very large input is passed.
    # Hence we limit the size of input passed into PyMarian within one API call.
    SHARD_SIZE = 5000

    @classmethod
    def load_model(cls, model_name: str, gpu: bool = False):
        """See parent class docstring for details on functionality and arguments."""
        repo_id = cls.MODEL_NAME_TO_HF_PATH[model_name]
        model_path = hf_hub_download(repo_id, filename="checkpoints/marian.model.bin")
        vocab_path = hf_hub_download(repo_id, filename="vocab.spm")
        marian_args = f"-m {model_path} -v {vocab_path} {vocab_path} --like comet-qe"
        if gpu:
            marian_args += cls.MARIAN_GPU_ARGS
        else:
            marian_args += cls.MARIAN_CPU_ARGS
        return cls(model_name, pymarian.Evaluator(marian_args), gpu)

    @staticmethod
    def wrap_qe_input(src: str, tgt: str, reverse=False, **kwargs):
        """See parent class docstring for details on functionality and arguments."""
        return [src, tgt] if not reverse else [tgt, src]

    def predict(self, input: List, **kwargs) -> List[float]:
        """Implements quality estimation score prediction for Cometoid/PyMarian model.

        Args:
            input (List): A list of bitext pairs wrapped as dictionaries.

        Returns:
            List[float]: List of quality scores.
        """
        scores = []
        for start_idx in range(0, len(input), self.SHARD_SIZE):
            scores.extend(
                self._model.evaluate(input[start_idx : start_idx + self.SHARD_SIZE])
            )

        if not self._name.endswith("mqm"):
            # using DA+SQM score by default
            # choice made based on paper: https://aclanthology.org/2023.wmt-1.62.pdf
            return [score[1] for score in scores]
        else:
            return [score[0] for score in scores]



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
from typing import Union

import cudf
import dask_cudf
import numpy as np
from dask.utils import M

from nemo_curator.log import create_logger
from nemo_curator.utils.fuzzy_dedup_utils.output_map_utils import (
    build_partition,
    get_agg_text_bytes_df,
)


class _MapBuckets:
    """
    buckets to a logical partition by using a modified bin packing algorithm.
    Combines buckets generated from LSH (typically high cardinality)
    to more coarse lower cardinality bucket groups by mapping multiple buckets
    to a logical partition using document length information and a modified bin
    packing algorithm.
    Only needed if running False Postive check to remove false positives.
    """

    def __init__(
        self,
        id_fields: Union[list, str] = "id",
        text_field: str = "text",
        bucket_field: str = "_bucket_id",
        num_anchors: int = 2,
        logger: Union[logging.LoggerAdapter, str] = "./",
    ):
        """
        id_fields: list or str
            id fields of df
        text_field: str = "text",
        bucket_column: str = "bucket_column",
        num_anchors: int = 2,
        logger: Union[logging.LoggerAdapter, str] = "./",
        """
        self.id_fields = [id_fields] if isinstance(id_fields, str) else id_fields
        self.text_field = text_field
        self.num_anchors = num_anchors
        self.bucket_field = bucket_field
        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "Map_Buckets.log"),
                name="Map_Buckets",
            )
        else:
            self._logger = logger

    @staticmethod
    def _get_output_part_ids_with_approx_equal_sum(
        bucket_text_bytes_df: cudf.DataFrame,
        max_text_bytes_per_part: int,
        buckets_column: str,
        bytes_column: str,
        output_partition_column: str,
    ) -> cudf.DataFrame:
        """
        Create a output_series that maps the ser.index into `nparts`
        so that the total sum of bucket_val_counts_df
        for each output id are all most equal and
        less than max_text_bytes_per_part
        This is used downstream for creating equal output_ids
        """
        sizes = bucket_text_bytes_df[bytes_column].values
        bucket_output_ar = build_partition(
            sizes=sizes.get(), max_size=max_text_bytes_per_part
        )
        df = cudf.DataFrame()
        df[buckets_column] = bucket_text_bytes_df[buckets_column]
        df[output_partition_column] = bucket_output_ar
        return df

    def _get_output_map_from_text_bytes_per_bucket(
        self,
        ddf_bk_text_bytes,
        bytes_column,
        output_partition_column="_output_partition_id",
    ):
        # String bytes limit for cuDF
        # https://github.com/rapidsai/cudf/issues/13733
        max_text_bytes_per_part = int(np.iinfo(np.int32).max * 3)

        self._logger.info(f"max_text_bytes_per_part = {max_text_bytes_per_part}")
        # Increasing in an attempt to prevent hitting
        # ulimits
        output_map_df_meta = cudf.DataFrame(
            {self.bucket_field: [0], output_partition_column: [1]}
        )
        output_map_df_meta = output_map_df_meta.astype(
            {self.bucket_field: np.uint64, output_partition_column: np.int32}
        )

        output_map_df = ddf_bk_text_bytes.map_partitions(
            _MapBuckets._get_output_part_ids_with_approx_equal_sum,
            max_text_bytes_per_part=max_text_bytes_per_part,
            buckets_column=self.bucket_field,
            bytes_column=bytes_column,
            output_partition_column=output_partition_column,
            meta=output_map_df_meta,
        )
        output_map_df = output_map_df.persist()
        self._logger.info(
            f"Step 1 of output_map_df of len: {len(output_map_df)} computed"
        )
        lower_bounds = (
            output_map_df[output_partition_column]
            .map_partitions(lambda s: (s.max() + 1))
            .compute()
        )
        lower_bounds = np.cumsum(lower_bounds)

        def update_id(df, lower_bound):
            df[output_partition_column] += lower_bound
            return df

        updated_parts = [
            output_map_df.get_partition(i).map_partitions(
                update_id, lower_bounds[i - 1]
            )
            for i in range(1, len(lower_bounds))
        ]
        updated_parts.append(output_map_df.get_partition(0))
        output_map_df = dask_cudf.concat(updated_parts)
        output_map_df = output_map_df.persist()
        self._logger.info(
            f"All steps of output_map_df of len: {len(output_map_df)} computed"
        )
        return output_map_df

    def _get_output_map_based_on_str_bytes(
        self, buckets_df, documents_df, bytes_column="_text_bytes"
    ):
        """
        Add output_partition_id to buckets_ddf
        """
        documents_df = documents_df.copy()
        documents_df[bytes_column] = documents_df[self.text_field].map_partitions(
            lambda s: s.str.byte_count()
        )
        n_partitions = buckets_df.npartitions
        documents_df = documents_df.drop(columns=[self.text_field]).repartition(
            npartitions=n_partitions
        )
        buckets_df = buckets_df.merge(documents_df).repartition(
            npartitions=n_partitions
        )
        del documents_df
        ddf_bk_text_bytes, agg_df_len = get_agg_text_bytes_df(
            df=buckets_df,
            agg_column=self.bucket_field,
            bytes_column=bytes_column,
            n_partitions=n_partitions,
            shuffle=True,
        )
        self._logger.info(f"Agg_df computed of length = {agg_df_len}")
        del buckets_df
        output_map_df = self._get_output_map_from_text_bytes_per_bucket(
            ddf_bk_text_bytes=ddf_bk_text_bytes,
            bytes_column=bytes_column,
        )
        return output_map_df

    def _random_select_anchor(self, buckets_df, n=2):
        """
        Randomly select `n` anchors from each bucket.
        """
        buckets_df = buckets_df.copy()
        buckets_df["_id_hash"] = buckets_df[self.id_fields].hash_values()
        buckets_df = buckets_df.sort_values([self.bucket_field, "_id_hash"])
        buckets_df["_order_in_bucket"] = buckets_df.groupby(
            self.bucket_field
        ).cumcount()
        buckets_df["is_anchor"] = buckets_df["_order_in_bucket"] < n
        for i in range(0, n):
            buckets_df[f"is_anchor_id_{i}"] = buckets_df["_order_in_bucket"] == i
        buckets_df = buckets_df.drop(columns=["_id_hash", "_order_in_bucket"], axis=1)
        buckets_df = buckets_df.reset_index(drop=True)
        buckets_df = buckets_df[buckets_df.is_anchor]
        return buckets_df

    def _add_anchor_docs(self, buckets_df, num_anchors):
        """
        Get anchor documents for each bucket.
        """
        df_anchor_bk = self._random_select_anchor(buckets_df=buckets_df, n=num_anchors)
        df_anchor_docs = None
        for i in range(num_anchors):
            df_anchor_bk_i = df_anchor_bk[df_anchor_bk[f"is_anchor_id_{i}"]][
                [self.bucket_field] + self.id_fields
            ].reset_index(drop=True)
            column_mapping = {id: f"anchor_{i}_{id}" for id in self.id_fields}
            df_anchor_bk_i = df_anchor_bk_i.rename(columns=column_mapping)
            if i == 0:
                df_anchor_docs = df_anchor_bk_i
            else:
                df_anchor_docs = df_anchor_bk_i.merge(
                    df_anchor_docs, on=[self.bucket_field], how="inner"
                )

        df_anchor_docs_with_bk = buckets_df.merge(
            df_anchor_docs, on=[self.bucket_field], how="inner"
        )
        return df_anchor_docs_with_bk

    def map_buckets_with_anchors(
        self,
        documents_df: dask_cudf.DataFrame,
        buckets_df: dask_cudf.DataFrame,
        shuffle_type: Union[str, bool, None] = "tasks",
    ) -> dask_cudf.DataFrame:
        """
        Get anchor docs with bucket info
        Args:
            input_data_paths: list of paths to input data
            input_bucket_path: path to input buckets
            text_ddf_blocksize: blocksize for text ddf
            num_files: number of files to read
            num_workers: number of workers
            shuffle_type: type of shuffle to use
        Returns:
            ddf_anchor_docs_with_bk
        """
        output_map_df = self._get_output_map_based_on_str_bytes(
            buckets_df=buckets_df, documents_df=documents_df
        )
        ddf_anchor_docs_with_bk = buckets_df.map_partitions(
            self._add_anchor_docs, num_anchors=self.num_anchors
        )
        self._logger.info("output_map_df is based on string bytes")
        ddf_anchor_docs_with_bk = ddf_anchor_docs_with_bk.merge(
            output_map_df, on=self.bucket_field
        )
        # Bucket is no longer needed
        ddf_anchor_docs_with_bk = ddf_anchor_docs_with_bk.drop(
            columns=[self.bucket_field]
        )
        # Below removes any duplicates lying around after dropping buckets
        ddf_anchor_docs_with_bk = ddf_anchor_docs_with_bk.map_partitions(
            M.drop_duplicates,
            meta=ddf_anchor_docs_with_bk._meta,
            enforce_metadata=False,
            transform_divisions=False,
            align_dataframes=False,
        )
        ddf_anchor_docs_with_bk = ddf_anchor_docs_with_bk.shuffle(
            self.id_fields,
            ignore_index=True,
            shuffle_method=shuffle_type,
        ).map_partitions(
            M.drop_duplicates,
            meta=ddf_anchor_docs_with_bk._meta,
            enforce_metadata=False,
            transform_divisions=False,
            align_dataframes=False,
        )
        del output_map_df
        return ddf_anchor_docs_with_bk



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
import os
import time
from typing import Optional, Union

import cudf
import cugraph.dask as dcg
import cugraph.dask.comms.comms as Comms
import cupy as cp
import dask_cudf
import numpy as np
from cugraph import MultiGraph
from dask.utils import M

from nemo_curator.log import create_logger
from nemo_curator.utils.distributed_utils import performance_report_if_with_ts_suffix


class ConnectedComponents:
    def __init__(
        self,
        cache_dir: str,
        jaccard_pairs_path: str,
        id_column="id",
        jaccard_threshold: float = 0.8,
        logger: Union[logging.LoggerAdapter, str] = "./",
        profile_dir: Optional[str] = None,
    ):
        self.cache_dir = cache_dir
        self.jaccard_pairs_path = jaccard_pairs_path
        self.id_column = id_column
        self.left_id = f"{id_column}_x"
        self.right_id = f"{id_column}_y"
        self.jaccard_threshold = jaccard_threshold
        self.profile_dir = profile_dir
        if isinstance(logger, str):
            self._logger = create_logger(
                rank=0,
                log_file=os.path.join(logger, "ConnectedComponents.log"),
                name="ConnectedComponents",
            )
        else:
            self._logger = logger

    def cc_workflow(self, output_path):
        deduped_parsed_id_path = self._write_dedup_parsed_id()
        encoded_jaccard_pair_path = self._write_encoded_jaccard_pair(
            deduped_parsed_id_path
        )
        deduped_encoded_jaccard_path = self._write_dedup_encoded_jaccard_pair(
            encoded_jaccard_pair_path
        )
        cc_path = self._run_connected_components(
            deduped_encoded_jaccard_path, deduped_parsed_id_path, output_path
        )
        return cc_path

    def _run_connected_components(
        self,
        deduped_encoded_jaccard_path,
        deduped_parsed_id_path,
        output_path,
    ):
        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir, "connected-components-run"
        ):

            Comms.initialize(p2p=False)
            df = dask_cudf.read_parquet(
                deduped_encoded_jaccard_path, blocksize="1GB", aggregate_files=True
            )
            df = df[df["jaccard"] == 1].reset_index(drop=True)

            labels_df = dask_cudf.read_parquet(deduped_parsed_id_path)
            num_nodes = len(labels_df)
            self_edge_df = labels_df[["uid"]].rename(columns={"uid": self.left_id})
            self_edge_df[self.right_id] = self_edge_df[self.left_id]

            df = df[[self.left_id, self.right_id]].astype(np.int64)
            df = dask_cudf.concat([df, self_edge_df])

            G = MultiGraph(directed=False)
            G.from_dask_cudf_edgelist(
                df, source=self.left_id, destination=self.right_id, renumber=False
            )
            result = dcg.weakly_connected_components(G)
            del G
            max_partitions = min(32, result.npartitions)
            n_components = len(
                result[["labels"]].drop_duplicates(split_out=max_partitions)
            )
            num_labels = len(result)
            labels_df = labels_df.merge(
                result, left_on=["uid"], right_on=["vertex"], how="inner"
            )
            id_columns = [self.id_column]
            labels_df = labels_df[id_columns + ["labels"]]
            labels_df = labels_df.rename(columns={"labels": "group"})
            labels_df = labels_df.persist()
            # Doing an inner merge above
            # should not change any rows

            self._logger.info(
                "Result of connected compoinents are "
                f"# of groups : {n_components}, "
                f"# of docs removed : {num_labels - n_components}, "
                f"# nodes = {num_nodes}, "
                f"# rows in labels_df = {len(labels_df)}"
            )
            assert num_nodes == len(labels_df)
            # Ensure all docs in the same group are in the same partition
            labels_df = labels_df.shuffle(on=["group"], ignore_index=True)
            labels_df.to_parquet(output_path, write_index=False, overwrite=True)
            Comms.destroy()
        self._logger.info(
            f"Time taken for Connected Components Run = {time.time() - t0}s and output written at {output_path}"
        )

    @staticmethod
    def _sort_ids(df, id_columns):
        x = df[id_columns].values
        x = cp.sort(x, axis=1)
        for i, id_column in enumerate(id_columns):
            df[id_column] = x[:, i]
            df[id_column] = df[id_column].astype("uint64")
        return df

    @staticmethod
    def thresholding(df, threshold, column_to_threshold):
        mask = df[column_to_threshold] > threshold
        df.loc[mask, column_to_threshold] = np.int8(1)
        df.loc[~mask, column_to_threshold] = np.int8(0)
        return df

    def _write_dedup_encoded_jaccard_pair(self, encoded_jaccard_pair_path):
        output_path = f"{self.cache_dir}/final_dedup_encoded_jaccard_pair.parquet"
        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir, "connected-components-dedup-encoded-jaccard-pair"
        ):

            ddf = dask_cudf.read_parquet(
                encoded_jaccard_pair_path, blocksize="512MB", aggregate_files=True
            )
            meta = {
                self.left_id: "uint64",
                self.right_id: "uint64",
                "jaccard": "float32",
            }
            ddf = ddf.map_partitions(
                ConnectedComponents._sort_ids,
                id_columns=[self.left_id, self.right_id],
                meta=meta,
            )
            ddf = ddf.map_partitions(
                ConnectedComponents.thresholding,
                threshold=self.jaccard_threshold,
                column_to_threshold="jaccard",
                meta=meta,
            )
            ddf = ddf.map_partitions(
                M.drop_duplicates,
                meta=ddf._meta,
                enforce_metadata=False,
                transform_divisions=False,
                align_dataframes=False,
            )

            ddf = ddf.shuffle(
                [self.left_id, self.right_id],
                ignore_index=True,
                shuffle_method="tasks",
            )
            ddf = ddf.map_partitions(
                M.drop_duplicates,
                meta=ddf._meta,
                enforce_metadata=False,
                transform_divisions=False,
                align_dataframes=False,
            )
            ddf.to_parquet(output_path, write_index=False, overwrite=True)
        self._logger.info(
            f"Time taken for Dedup Encoding Jaccard Pairs = {time.time() - t0}s and output written at {output_path}"
        )
        return output_path

    def _write_dedup_parsed_id(self):
        dedup_parsed_id_path = f"{self.cache_dir}/dedup_parsed_id.parquet"
        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir, "connected-components-dedup-parsed-id"
        ):
            ddf = dask_cudf.read_parquet(
                self.jaccard_pairs_path,
                columns=[self.left_id, self.right_id],
                blocksize="512MB",
                aggregate_files=True,
            )
            id_columns = [self.id_column]
            unique_docs = ddf.map_partitions(
                ConnectedComponents._get_unique_ids_per_partition, id_columns=id_columns
            )
            unique_docs = unique_docs.drop_duplicates(
                # Dask does not guard against split_out=0
                split_out=max(ddf.npartitions // 4, 1)
            )
            unique_docs["uid"] = np.uint64(1)
            unique_docs["uid"] = unique_docs["uid"].cumsum()
            unique_docs["uid"] = unique_docs["uid"] - 1
            unique_docs.to_parquet(
                dedup_parsed_id_path, write_index=False, overwrite=True
            )
        self._logger.info(
            f"Time taken for Dedup Parsed Id = {time.time() - t0}s and output written at {dedup_parsed_id_path}"
        )
        return dedup_parsed_id_path

    def _write_encoded_jaccard_pair(self, dedup_parsed_id_path):
        output_path = f"{self.cache_dir}/encoded_jaccard_pair/"
        t0 = time.time()
        with performance_report_if_with_ts_suffix(
            self.profile_dir, "connected-components-encoded-jaccard-pair"
        ):
            ddf_id = dask_cudf.read_parquet(
                dedup_parsed_id_path, blocksize="2GB", aggregate_files=True
            )
            ddf = dask_cudf.read_parquet(
                self.jaccard_pairs_path,
                blocksize="1GB",
                aggregate_files=True,
            )
            self._merge_and_write(
                ddf=ddf,
                ddf_id=ddf_id,
                output_path=output_path,
                id_column=self.id_column,
            )
        self._logger.info(
            f"Time taken for Encoding Jaccard Pairs = {time.time() - t0}s and output written at {output_path}"
        )
        return output_path

    def _merge_and_write(
        self,
        ddf: dask_cudf.DataFrame,
        ddf_id: dask_cudf.DataFrame,
        output_path: str,
        id_column: str,
    ) -> None:
        st = time.time()
        # Ensure 'id_columns' is a list
        ddf_id = ddf_id.set_index(id_column)
        for tag in ["x", "y"]:
            pair_id = f"{id_column}_{tag}"
            # Merge 'ddf' with 'ddf_id' to map ids to uids
            ddf = ddf.merge(
                ddf_id,
                left_on=pair_id,
                right_index=True,
                how="inner",
                broadcast=True,
            )
            ddf = ddf.drop(columns=pair_id)
            ddf = ddf.rename(columns={"uid": f"{self.id_column}_{tag}"})
        ddf = ddf[[self.left_id, self.right_id, "jaccard"]]
        ddf.to_parquet(output_path, write_index=False, overwrite=True)

        et = time.time()
        self._logger.info(
            f"Time taken for merge and write = {et - st}s and output written at {output_path}"
        )

    @staticmethod
    def _get_unique_ids_per_partition(df, id_columns):
        unique_df_ls = []
        for tag in ["x", "y"]:
            cols_to_drop = []
            for id_col in id_columns:
                cols_to_drop.append(f"{id_col}_{tag}")

            subset_df = df[cols_to_drop].drop_duplicates(ignore_index=True)
            subset_df = subset_df.rename(
                columns={f"{id_col}_{tag}": f"{id_col}" for id_col in id_columns}
            )
            unique_df_ls.append(subset_df)
        unique_df = cudf.concat(unique_df_ls, ignore_index=True)
        unique_df = unique_df.drop_duplicates(ignore_index=True)
        return unique_df



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import subprocess
import zlib
from collections import OrderedDict
from datetime import datetime, timedelta
from typing import List, Optional
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup


def get_main_warc_paths(
    snapshot_index,
    start_snapshot,
    end_snapshot,
    prefix="https://data.commoncrawl.org",
):
    beg_year, beg_week = list(map(int, start_snapshot.split("-")))
    end_year, end_week = list(map(int, end_snapshot.split("-")))
    start_date = datetime.fromisocalendar(beg_year, beg_week, 1)
    end_date = datetime.fromisocalendar(end_year, end_week, 1)

    if start_date > end_date:
        raise ValueError(
            f"Start snapshot '{start_snapshot}' is after end snapshot '{end_snapshot}'"
        )

    if beg_year < 2013 or end_year < 2013:
        print("Warning: Only snapshots after 2013 are supported by this script")

    total_prefix = urljoin(prefix, "crawl-data/CC-MAIN")

    warc_paths = []
    for snapshot in snapshot_index:
        date = list(map(int, snapshot["id"].split("-")[2:]))

        if len(date) == 2:
            year, week = date
        else:
            continue

        if year >= 2013:
            curr_date = datetime.fromisocalendar(year, week, 1)
            if curr_date >= start_date and curr_date <= end_date:
                warc_path = f"{total_prefix}-{year}-{week:02d}/warc.paths.gz"
                warc_paths.append(warc_path)

    return warc_paths


def get_news_warc_paths(
    start_date,
    end_date,
    prefix="https://data.commoncrawl.org",
):
    beg = datetime.strptime(start_date, "%Y-%m")
    end = datetime.strptime(end_date, "%Y-%m")

    # Get current year and month
    today = datetime.now()

    if start_date > end_date:
        raise ValueError(
            f"Start snapshot '{start_date}' is after end snapshot '{end_date}'"
        )

    if beg.year < 2016 or end.year > today.year:
        print(
            "Warning: WARC paths exist only from 2016-8 to "
            f"{today.year}-{today.month}"
        )
    total_prefix = urljoin(prefix, "crawl-data/CC-NEWS")

    # Generate all valid YYYY-MM strings in range
    dates = OrderedDict()
    for day in range((end - beg).days + 1):
        new_date = beg + timedelta(day)
        dates[(new_date.year, new_date.month)] = None

    dates = list(dates.keys())

    warc_paths = []
    for year, month in dates:
        warc_path = f"{total_prefix}/{year}/{month:02d}/warc.paths.gz"
        warc_paths.append(warc_path)

    return warc_paths


def get_common_crawl_snapshot_index(index_prefix):
    index_url = urljoin(index_prefix, "collinfo.json")
    index_response = requests.get(index_url)

    return json.loads(index_response.content)


def get_common_crawl_urls(
    starting_snapshot: str,
    ending_snapshot: str,
    data_domain_prefix="https://data.commoncrawl.org",
    index_prefix="https://index.commoncrawl.org",
    news=False,
) -> List[str]:
    """
    Retrieves the URLs for all the compressed WARC files between given Common Crawl snapshots

    Args:
    starting_snapshot: The first common crawl snapshot to include. Snapshots must be
        specified by YYYY-WeekNumber (e.g., '2020-50' or '2021-04'). For the CC-NEWS dataset,
        (specified with news=True flag) this changes to Year-Month (YYYY-MM).
    ending_snapshot: The last common crawl snapshot to include. Must be chronologically
        after the starting snapshot.
    data_domain_prefix: The prefix that will be prepended to each WARC file to create the URL.
    index_prefix: The prefix of the URL to the Common Crawl index.
    news: If True, gets WARC URLs for the CC-NEWS dataset instead of the CC-MAIN datasets.
        Also assumes that the format for the start and end snapshots is 'YYYY-MM' (Year-Month).
    """
    if news:
        warc_paths = get_news_warc_paths(
            starting_snapshot, ending_snapshot, prefix=data_domain_prefix
        )
    else:
        index = get_common_crawl_snapshot_index(index_prefix)
        warc_paths = get_main_warc_paths(
            index, starting_snapshot, ending_snapshot, prefix=data_domain_prefix
        )

    common_crawl_urls = []
    for path in warc_paths:
        try:
            response = requests.get(path.rstrip(), stream=True)
            data = zlib.decompress(response.content, zlib.MAX_WBITS | 32)
            for warc in data.decode("utf-8").split("\n"):
                if warc != "":
                    warc_url = urljoin(data_domain_prefix, warc)
                    common_crawl_urls.append(warc_url)
        except Exception as e:
            print(f"Could not get URLs for snapshot {path}")
            print(response.content)
            print(e)

    return common_crawl_urls


def get_wikipedia_urls(
    language="en",
    wikidumps_index_prefix="https://dumps.wikimedia.org",
    dump_date: Optional[str] = None,
) -> List[str]:
    """
    Retrieves all urls pointing to the latest Wikipedia dumps

    Args:
        language: Desired language of the Wikipedia dump.
        wikidumps_index_prefix: The base url for all wikipedia dumps
        dump_date: A string formatted as "YYYYMMDD" for the wikipedia dump to use.
          If None, latest dump is used.
    """
    wiki_index_url = urljoin(wikidumps_index_prefix, f"{language}wiki")
    if not dump_date:
        # First get the index
        raw_wiki_index = requests.get(wiki_index_url)
        wiki_index = raw_wiki_index.content.decode("utf-8")
        wiki_index_parsed = BeautifulSoup(wiki_index, "lxml")

        # Get all dumps available in the index
        dumps = wiki_index_parsed.find_all("a")
        dump_date = dumps[-2].text
    else:
        # A trailing / is needed for the url
        dump_date = dump_date + "/"

    # Get the json dump data
    wiki_latest_dump = urljoin(wiki_index_url + "/", dump_date)
    wiki_latest_dump_status = urljoin(wiki_latest_dump, "dumpstatus.json")
    raw_dump_data = requests.get(wiki_latest_dump_status)
    try:
        dump_data = json.loads(raw_dump_data.content)
    except json.decoder.JSONDecodeError:
        raise ValueError(f"No wikipedia dump found for {dump_date[:-1]}")

    # Get all multistream files within the dump data
    wikipedia_urls = []
    for ifile in dump_data["jobs"]["articlesmultistreamdump"]["files"]:
        if "xml" in ifile:
            url = urljoin(wiki_latest_dump, ifile)
            wikipedia_urls.append(url)

    return wikipedia_urls


def get_arxiv_urls():
    command = "s5cmd --request-payer=requester ls s3://arxiv/src/ | grep '.tar'"
    result = subprocess.run(command, capture_output=True, text=True, shell=True)

    if result.returncode != 0:
        raise RuntimeError(f"Unable to get arxiv urls: {result.stderr}")

    urls = result.stdout.split()[3::4]
    urls.sort()

    return urls



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import dask
import numpy as np
import pandas as pd
import pytest
from dask import dataframe as dd
from dask.dataframe.utils import assert_eq

from nemo_curator.datasets import DocumentDataset
from nemo_curator.datasets.parallel_dataset import ParallelDataset
from nemo_curator.filters import (
    AlphaFilter,
    BoilerPlateStringFilter,
    BulletsFilter,
    CommonEnglishWordsFilter,
    DocumentFilter,
    EllipsisFilter,
    GeneralCommentToCodeFilter,
    HistogramFilter,
    HTMLBoilerplateFilter,
    LengthRatioFilter,
    LongWordFilter,
    MeanWordLengthFilter,
    NonAlphaNumericFilter,
    NumberOfLinesOfCodeFilter,
    NumbersFilter,
    ParenthesesFilter,
    PerExtensionFilter,
    PornographicUrlsFilter,
    PunctuationFilter,
    PythonCommentToCodeFilter,
    RepeatedLinesByCharFilter,
    RepeatedLinesFilter,
    RepeatedParagraphsByCharFilter,
    RepeatedParagraphsFilter,
    RepeatingDuplicateNGramsFilter,
    RepeatingTopNGramsFilter,
    SubstringFilter,
    SymbolsToWordsFilter,
    TokenCountFilter,
    TokenizerFertilityFilter,
    UrlsFilter,
    WhiteSpaceFilter,
    WordCountFilter,
    WordsWithoutAlphabetsFilter,
    XMLHeaderFilter,
)
from nemo_curator.filters.models.qe_models import COMET_IMPORT_MSG, PYMARIAN_IMPORT_MSG
from nemo_curator.modules import (
    Filter,
    ParallelScoreFilter,
    Score,
    ScoreFilter,
    Sequential,
)
from nemo_curator.utils.decorators import batched
from nemo_curator.utils.import_utils import is_unavailable, safe_import

comet = safe_import("comet", msg=COMET_IMPORT_MSG)
pymarian = safe_import("pymarian", msg=PYMARIAN_IMPORT_MSG)


class LetterCountFilter(DocumentFilter):
    """
    Keeps documents that have at least some number of a given letter
    """

    def __init__(self, letter="a", min_count=5):
        super().__init__()
        self.letter = letter
        self.min_count = min_count

    def score_document(self, text):
        return text.count(self.letter)

    def keep_document(self, score):
        return score >= self.min_count


class BatchedLengthFilter(DocumentFilter):
    """
    Keeps documents of a given length
    """

    def __init__(self, min_length=5, max_length=10):
        super().__init__()
        self.min_length = min_length
        self.max_length = max_length

    @batched
    def score_document(self, df):
        return df.str.len()

    @batched
    def keep_document(self, scores):
        min_threshold = self.min_length <= scores
        max_threshold = scores <= self.max_length
        return min_threshold & max_threshold


# A simple dummy tokenizer for our tests.
class DummyTokenizer:
    def encode(self, text):
        # Simply splits the text on whitespace.
        return text.split()


def all_equal(left_dataset, right_dataset):
    return all(left_dataset.df.compute() == right_dataset.df.compute())


def list_to_dataset(documents, col_name="text", npartitions=2):
    data = {col_name: documents}
    pdf = pd.DataFrame(data)

    return DocumentDataset(dd.from_pandas(pdf, npartitions=npartitions))


def two_lists_to_parallel_dataset(
    src_documents,
    tgt_documents,
    src_lang,
    tgt_lang,
    src_col_name="src",
    tgt_col_name="tgt",
    npartitions=2,
):
    src_langs = [src_lang] * len(src_documents)
    tgt_langs = [tgt_lang] * len(src_documents)
    data = {
        src_col_name: src_documents,
        "src_lang": src_langs,
        tgt_col_name: tgt_documents,
        "tgt_lang": tgt_langs,
    }
    pdf = pd.DataFrame(data)

    return ParallelDataset(dd.from_pandas(pdf, npartitions=npartitions))


@pytest.fixture
def letter_count_data():
    return list_to_dataset(
        ["Two aa", "a a Three a", "Five aaa aa", "aaaSeven aaaa"], col_name="documents"
    )


@pytest.fixture
def parallel_letter_count_data():
    return two_lists_to_parallel_dataset(
        ["Einsa", "Zwei aaa", "a Drei a", "Fünf aaa a", "aaaSieben aaaa"],
        ["aOne", "Two aa", "a a Three a", "Five aaa aa", "aaaSeven aaaa"],
        src_lang="de",
        tgt_lang="en",
        src_col_name="src",
        tgt_col_name="tgt",
    )


@pytest.fixture
def length_ratio_data():
    return two_lists_to_parallel_dataset(
        ["Test", "test", "Test Test ", "Test Test"],
        ["Prueba", "prueba prueba prueba", "Prueba Prueba", "Prueba Prueba Prueba "],
        src_lang="en",
        tgt_lang="es",
    )


class TestFilterModule:
    def test_score_filter(self, letter_count_data):
        letter_filter = LetterCountFilter()
        filter_step = ScoreFilter(letter_filter, text_field="documents")
        filtered_data = filter_step(letter_count_data)

        expected_indices = [2, 3]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_score(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter.score_document,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([2, 3, 5, 7])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_score_document(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([2, 3, 5, 7])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_retain_score_filter(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "count_a"
        filter_step = ScoreFilter(
            letter_filter, text_field="documents", score_field=score_field
        )
        filtered_data = filter_step(letter_count_data)

        expected_indices = [2, 3]
        # Compute before loc due to https://github.com/dask/dask-expr/issues/1036
        expected_data = letter_count_data.df.compute().loc[expected_indices]
        expected_data = DocumentDataset(dd.from_pandas(expected_data, 2))
        expected_data.df[score_field] = pd.Series([5, 7], index=expected_data.df.index)
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_filter(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter.score_document,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)
        filter_step = Filter(letter_filter.keep_document, score_field)
        filtered_data = filter_step(scored_data)

        expected_indices = [2, 3]
        # Compute before loc due to https://github.com/dask/dask-expr/issues/1036
        expected_data = letter_count_data.df.compute().loc[expected_indices]
        expected_data = dd.from_pandas(expected_data, 2)
        expected_data[score_field] = pd.Series([5, 7], index=expected_data.index)
        expected_data = DocumentDataset(expected_data)
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_filter_document(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)
        filter_step = Filter(letter_filter, score_field)
        filtered_data = filter_step(scored_data)

        expected_indices = [2, 3]
        # Compute before loc due to https://github.com/dask/dask-expr/issues/1036
        expected_data = letter_count_data.df.compute().loc[expected_indices]
        expected_data = dd.from_pandas(expected_data, 2)
        expected_data[score_field] = pd.Series([5, 7], index=expected_data.index)
        expected_data = DocumentDataset(expected_data)
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_invert(self, letter_count_data):
        letter_filter = LetterCountFilter()
        filter_step = ScoreFilter(letter_filter, text_field="documents", invert=True)
        filtered_data = filter_step(letter_count_data)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_sequential_filter(self, letter_count_data):
        filters = Sequential(
            [
                ScoreFilter(LetterCountFilter(), text_field="documents"),
                ScoreFilter(LetterCountFilter(min_count=6), text_field="documents"),
            ]
        )
        filtered_data = filters(letter_count_data)

        expected_indices = [3]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_batch_score_filter(self, letter_count_data):
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        filter_step = ScoreFilter(length_filter, text_field="documents")
        filtered_data = filter_step(letter_count_data)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_batch_score(self, letter_count_data):
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        score_field = "lengths"
        score_step = Score(
            length_filter.score_document,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([6, 11, 11, 13])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_batch_score_document(self, letter_count_data):
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        score_field = "lengths"
        score_step = Score(
            length_filter,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([6, 11, 11, 13])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_batch_filter(self, letter_count_data):
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        score_field = "lengths"
        score_step = Score(
            length_filter.score_document,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)
        filter_step = Filter(length_filter.keep_document, score_field)
        filtered_data = filter_step(scored_data)

        expected_indices = [1, 2]
        expected_data = letter_count_data.df.loc[expected_indices]
        expected_data[score_field] = pd.Series([11, 11], index=expected_data.index)
        expected_data = DocumentDataset(expected_data)
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_batch_filter_document(self, letter_count_data):
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        score_field = "lengths"
        score_step = Score(
            length_filter,
            text_field="documents",
            score_field=score_field,
        )
        scored_data = score_step(letter_count_data)
        filter_step = Filter(length_filter, score_field)
        filtered_data = filter_step(scored_data)

        expected_indices = [1, 2]
        expected_data = letter_count_data.df.loc[expected_indices]
        expected_data[score_field] = pd.Series([11, 11], index=expected_data.index)
        expected_data = DocumentDataset(expected_data)
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_score_filter_type(self, letter_count_data):
        letter_filter = LetterCountFilter()
        filter_step = ScoreFilter(letter_filter, text_field="documents", score_type=int)
        filtered_data = filter_step(letter_count_data)

        expected_indices = [2, 3]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_score_type(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter.score_document,
            text_field="documents",
            score_field=score_field,
            score_type=int,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([2, 3, 5, 7])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_score_type_document(self, letter_count_data):
        letter_filter = LetterCountFilter()
        score_field = "a_count"
        score_step = Score(
            letter_filter,
            text_field="documents",
            score_field=score_field,
            score_type=int,
        )
        scored_data = score_step(letter_count_data)

        expected_scores = pd.Series([2, 3, 5, 7])
        scores = scored_data.df[score_field]
        assert all(
            expected_scores == scores.compute()
        ), f"Expected {expected_scores} but got {scores}"

    def test_chain_filter(self, letter_count_data):
        letter_count_filter = LetterCountFilter(min_count=4)
        length_filter = BatchedLengthFilter(min_length=8, max_length=11)
        filters = Sequential(
            [
                ScoreFilter(letter_count_filter, text_field="documents"),
                ScoreFilter(length_filter, text_field="documents"),
            ]
        )
        filtered_data = filters(letter_count_data)

        expected_indices = [2]
        expected_data = DocumentDataset(letter_count_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_parallel_score_filter(self, parallel_letter_count_data):
        src_letter_count_filter = LetterCountFilter(min_count=2)
        tgt_letter_count_filter = LetterCountFilter(min_count=3)
        filter_step = ParallelScoreFilter(
            src_letter_count_filter, tgt_letter_count_filter
        )
        filtered_data = filter_step(parallel_letter_count_data)

        expected_indices = [2, 3, 4]
        expected_data = ParallelDataset(
            parallel_letter_count_data.df.loc[expected_indices]
        )
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_joint_score_filter(self, length_ratio_data):
        filter_ = LengthRatioFilter(
            max_ratio=1.5,
            src_lang="en",
            tgt_lang="de",
            score_field="ratio",
            score_type=float,
        )
        filtered_data = filter_(length_ratio_data)

        expected_indices = [0, 2]
        expected_data = ParallelDataset(length_ratio_data.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"


class TestHeuristicFilters:
    def test_nonalpha(self):
        dataset = list_to_dataset(
            ["", "This is a test case.", "%$^%$^%$&^$()))))", "$aaa"]
        )
        filters = ScoreFilter(NonAlphaNumericFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_symbolswords(self):
        dataset = list_to_dataset(
            [
                "mixed bag ... #",
                "full of words",
                "... # ... # #",
                "barely ok 3 4 5 6 7 8 9 #",
            ]
        )
        filters = ScoreFilter(SymbolsToWordsFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_numbers(self):
        dataset = list_to_dataset(
            ["purely letters", "34134543", "$!@$@!$!@", "abcdefghi1"]
        )
        filters = ScoreFilter(NumbersFilter(max_number_to_text_ratio=0.1))
        filtered_data = filters(dataset)

        expected_indices = [0, 2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_urls(self):
        dataset = list_to_dataset(
            [
                "https://www.nvidia.com/en-us/",
                "no urls here!",
                "$!@$@!$!@",
                "bunch of other words with url afdsjafidsaofjbwreowihfdsafbdashuoiotauhiofdafdsafd fdasfdafdsafdsafdsafdsafdsafdsa https://www.nvidia.com/en-us/ something else after the url etc more and more",
                "words with url https://www.nvidia.com/en-us/",
            ]
        )
        filters = ScoreFilter(UrlsFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_bullets(self):
        dataset = list_to_dataset(
            [
                "• not good",
                "good",
                "50 \n ⦾ 50",
                "⁌ this \n⁌ should \n⁌barely \n⁌pass \n⁌5 \n⁌6 \n⁌7 \n⁌8 \n⁌9 \n done!",
            ]
        )
        filters = ScoreFilter(BulletsFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_whitespace(self):
        dataset = list_to_dataset(["\t\n\r", "good", "50%\n\n\n", "123\b"])
        filters = ScoreFilter(WhiteSpaceFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_parentheses(self):
        dataset = list_to_dataset(
            ["()", "(not good)", "this is completely absolutely fine", "123456789("]
        )
        filters = ScoreFilter(ParenthesesFilter())
        filtered_data = filters(dataset)

        expected_indices = [2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_longword(self):
        dataset = list_to_dataset(["tiny", "large"])
        filters = ScoreFilter(LongWordFilter(max_word_length=4))
        filtered_data = filters(dataset)

        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_wordcount(self):
        dataset = list_to_dataset(
            ["", "one", "two words", "$#@$ %$@$#@ !#@!", "one two three four five"]
        )
        filters = ScoreFilter(WordCountFilter(min_words=2, max_words=4))
        filtered_data = filters(dataset)

        expected_indices = [2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_wordcount_zh(self):
        dataset = list_to_dataset(["", "你好。", "我喜欢学习中文。"])
        filters = ScoreFilter(WordCountFilter(min_words=2, max_words=5, lang="zh"))
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_wordcount_ja(self):
        dataset = list_to_dataset(
            ["", "猫が寝ます。", "私は日本語のテキストを分割します。"]
        )
        filters = ScoreFilter(WordCountFilter(min_words=5, max_words=11, lang="ja"))
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_boilerplate(self):
        dataset = list_to_dataset(
            [
                "nothing\t here",
                "1\n\n2\n\n3\n\n4\n\n5\n\n6\n\nterms of use\n\n privacy policy\n\n cookie policy\n\nuses cookies",
                "too much \n\n privacy & cookies policy",
            ]
        )
        filters = ScoreFilter(BoilerPlateStringFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_meanwordlength(self):
        dataset = list_to_dataset(
            [
                "a",
                "aa",
                "superlongword short",
                "evenly balanced",
                "waytoolongforasingleword",
            ]
        )
        filters = ScoreFilter(MeanWordLengthFilter())
        filtered_data = filters(dataset)

        expected_indices = [2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatedlines(self):
        dataset = list_to_dataset(["totally unique", "half.\nhalf."])
        filters = ScoreFilter(RepeatedLinesFilter())
        filtered_data = filters(dataset)

        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatedparagraphs(self):
        dataset = list_to_dataset(["totally unique", "half.\n\nhalf."])
        filters = ScoreFilter(RepeatedParagraphsFilter())
        filtered_data = filters(dataset)

        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatedlineschar(self):
        dataset = list_to_dataset(
            [
                "totally unique",
                "a.\na.\nvery very very short duplicate.",
                "half.\nhalf.",
                "super very incredibly huge long duplicate.\nsuper very incredibly huge long duplicate.\na.\nb.\nc.",
            ]
        )
        filters = ScoreFilter(RepeatedLinesByCharFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatedparagraphschar(self):
        dataset = list_to_dataset(
            [
                "totally unique",
                "a.\n\n  a.\n\n  very very very short duplicate.",
                "half.\n\nhalf.",
                "super very incredibly huge long duplicate.\n\nsuper very incredibly huge long duplicate.\n\n  a.\n\n  b.\n\n  c.",
            ]
        )
        filters = ScoreFilter(RepeatedParagraphsByCharFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatingtopngrams(self):
        dataset = list_to_dataset(
            [
                "this is a totally fine sentence with no repeat ngrams so we are ok",
                "a b . a b",
                "a a a a a a",
                "totally fine small dupe a b a b",
            ]
        )
        filters = ScoreFilter(RepeatingTopNGramsFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_repeatingduplicatengrams(self):
        dataset = list_to_dataset(
            ["a a b b a a b b", "totally fine", "a a a a this should be fine as well"]
        )
        filters = ScoreFilter(RepeatingDuplicateNGramsFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_punctuation(self):
        dataset = list_to_dataset(
            ["not good", "good.", "just\n barely\n fine\n ok\n yep."]
        )
        filters = ScoreFilter(
            PunctuationFilter(max_num_sentences_without_endmark_ratio=0.8)
        )
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_ellipsis(self):
        dataset = list_to_dataset(
            ["not good...", "good.", "just...\n barely...\n fine...\n ok...\n yep."]
        )
        filters = ScoreFilter(
            EllipsisFilter(max_num_lines_ending_with_ellipsis_ratio=0.8)
        )
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_commonenglishwords(self):
        dataset = list_to_dataset(["uncommon", "the and", "the and and of to"])
        filters = ScoreFilter(CommonEnglishWordsFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_wordswithoutalphabets(self):
        dataset = list_to_dataset(["totally fine", "good good good good !", "@"])
        filters = ScoreFilter(WordsWithoutAlphabetsFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_pornographicurls(self):
        dataset = list_to_dataset(
            [
                "no url",
                "fine url https://www.nvidia.com/en-us/",
                "bad url https://www.pornhub.com/",
            ]
        )
        filters = ScoreFilter(PornographicUrlsFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_histogram(self):
        dataset = list_to_dataset(
            [
                "This is a perfectly fine English document.",
                "But if you insist that this is written in Chinese,",
                "it's likely that something is fishy.",
                "另一方面，这是一个好的中文文档，",
                "但你一定要说这是英文文档，",
                "那很可能有些地方出了差错。",
            ]
        )
        filter1 = ScoreFilter(HistogramFilter(lang="en"))
        filter2 = ScoreFilter(HistogramFilter(lang="zh"))

        expected_indices1 = [0, 1, 2]
        expected_indices2 = [3, 4, 5]
        expected_data1 = DocumentDataset(dataset.df.loc[expected_indices1])
        expected_data2 = DocumentDataset(dataset.df.loc[expected_indices2])

        filtered_data1 = filter1(dataset)
        filtered_data2 = filter2(dataset)
        assert all_equal(
            expected_data1, filtered_data1
        ), f"Expected {expected_data1} but got {filtered_data1}"
        assert all_equal(
            expected_data2, filtered_data2
        ), f"Expected {expected_data2} but got {filtered_data2}"


class TestTokenCountFilter:
    def test_score_document(self):
        tokenizer = DummyTokenizer()
        token_filter = TokenCountFilter(tokenizer, min_tokens=2, max_tokens=3)
        text = "another test case"  # Should yield 3 tokens.
        score = token_filter.score_document(text)
        assert score == 3

    def test_keep_document(self):
        tokenizer = DummyTokenizer()
        token_filter = TokenCountFilter(tokenizer, min_tokens=2, max_tokens=3)
        # Check that a score of 1 (too few) and 4 (too many) are rejected,
        # while scores of 2 and 3 are accepted.
        assert token_filter.keep_document(2)
        assert token_filter.keep_document(3)
        assert not token_filter.keep_document(1)
        assert not token_filter.keep_document(4)

    def test_filter_dataset(self):
        # Create a dataset of documents with different word counts.
        docs = [
            "hello",  # 1 token
            "hello world",  # 2 tokens
            "this is a test",  # 4 tokens
            "another test case",  # 3 tokens
        ]
        dataset = list_to_dataset(docs, col_name="text")

        tokenizer = DummyTokenizer()
        token_filter = TokenCountFilter(tokenizer, min_tokens=2, max_tokens=3)
        filter_step = ScoreFilter(token_filter, text_field="text")
        filtered_dataset = filter_step(dataset)
        # Reset indices for filtered dataset to ensure identical labeling for comparison.
        filtered_dataset.df = filtered_dataset.df.reset_index(drop=True)

        # We expect to keep only the documents with exactly 2 or 3 tokens.
        expected_docs = [
            "hello world",  # 2 tokens
            "another test case",  # 3 tokens
        ]
        expected_dataset = list_to_dataset(expected_docs, col_name="text")
        # Reset indices for expected dataset to ensure identical labeling.
        expected_dataset.df = expected_dataset.df.reset_index(drop=True)
        assert all_equal(expected_dataset, filtered_dataset)

    def test_filter_dataset_default(self):
        # Create a dataset of documents with different word counts.
        docs = [
            "hello",  # 1 token
            "hello world",  # 2 tokens
            "this is a test",  # 4 tokens
            "another test case",  # 3 tokens
        ]
        dataset = list_to_dataset(docs, col_name="text")

        tokenizer = DummyTokenizer()
        # Using default settings: min_tokens=0 and max_tokens=inf, so all documents pass.
        token_filter = TokenCountFilter(tokenizer)
        filter_step = ScoreFilter(token_filter, text_field="text")
        filtered_dataset = filter_step(dataset)

        # We expect to keep all documents.
        expected_dataset = list_to_dataset(docs, col_name="text")
        assert all_equal(expected_dataset, filtered_dataset)


class TestSubstringFilter:
    def test_invalid_position(self):
        # Creating a SubstringFilter with an invalid position should raise a ValueError.
        with pytest.raises(ValueError):
            SubstringFilter("foo", "middle")

    def test_prefix_mode(self):
        filter_prefix = SubstringFilter("Hello", "prefix")
        # Positive example: text starts with "Hello".
        text = "Hello world"
        score = filter_prefix.score_document(text)
        assert score == 1
        assert filter_prefix.keep_document(score)
        # Negative example: text does not start with "Hello".
        text2 = "world Hello"
        score2 = filter_prefix.score_document(text2)
        assert score2 == 0
        assert not filter_prefix.keep_document(score2)

    def test_suffix_mode(self):
        filter_suffix = SubstringFilter("end", "suffix")
        # Positive example: text ends with "end".
        text = "This is the end"
        score = filter_suffix.score_document(text)
        assert score == 1
        assert filter_suffix.keep_document(score)
        # Negative example: text does not end with "end".
        text2 = "The end is near"
        score2 = filter_suffix.score_document(text2)
        assert score2 == 0
        assert not filter_suffix.keep_document(score2)

    def test_any_mode(self):
        filter_any = SubstringFilter("test", "any")
        # Positive example: text contains "test".
        text = "this is a test string"
        score = filter_any.score_document(text)
        assert score == 1
        assert filter_any.keep_document(score)
        # Negative example: text does not contain "test".
        text2 = "this is a string"
        score2 = filter_any.score_document(text2)
        assert score2 == 0
        assert not filter_any.keep_document(score2)

    def test_filter_dataset_prefix(self):
        docs = ["Hello world", "world Hello", "Hello everyone", "Not matching"]
        dataset = list_to_dataset(docs, col_name="text")
        filter_prefix = SubstringFilter("Hello", "prefix")
        filter_step = ScoreFilter(filter_prefix, text_field="text")
        filtered_dataset = filter_step(dataset)

        # Expect only those records where the text starts with "Hello".
        expected_docs = ["Hello world", "Hello everyone"]
        expected_dataset = list_to_dataset(expected_docs, col_name="text")

        # Reset indices to ensure both DataFrames are identically labeled
        filtered_dataset = DocumentDataset(filtered_dataset.df.reset_index(drop=True))
        expected_dataset = DocumentDataset(expected_dataset.df.reset_index(drop=True))
        assert all_equal(expected_dataset, filtered_dataset)

    def test_filter_dataset_suffix(self):
        docs = [
            "This is the end",  # ends with "end"
            "end of story",  # does not end with "end"
            "ending is good",  # does not end with "end"
            "Not matching end",  # ends with "end"
            "The end",  # ends with "end"
        ]
        dataset = list_to_dataset(docs, col_name="text")
        filter_suffix = SubstringFilter("end", "suffix")
        filter_step = ScoreFilter(filter_suffix, text_field="text")
        filtered_dataset = filter_step(dataset)

        # Expect only those records that end with "end".
        expected_docs = [
            "Not matching end",
            "The end",
            "This is the end",
        ]
        expected_dataset = list_to_dataset(expected_docs, col_name="text")

        # Compare only the 'text' column values to avoid index label issues.
        filtered_dataset = DocumentDataset(filtered_dataset.df.reset_index(drop=True))
        expected_dataset = DocumentDataset(expected_dataset.df.reset_index(drop=True))
        assert_eq(expected_dataset.df["text"], filtered_dataset.df["text"])

    def test_filter_dataset_any(self):
        docs = ["test case", "This is a testcase", "no match here", "another test"]
        dataset = list_to_dataset(docs, col_name="text")
        filter_any = SubstringFilter("test", "any")
        filter_step = ScoreFilter(filter_any, text_field="text")
        filtered_dataset = filter_step(dataset)

        # Expect documents that contain "test" anywhere.
        expected_docs = ["test case", "This is a testcase", "another test"]
        expected_dataset = list_to_dataset(expected_docs, col_name="text")

        # Reset indices to ensure both DataFrames are identically labeled
        filtered_dataset = DocumentDataset(filtered_dataset.df.reset_index(drop=True))
        expected_dataset = DocumentDataset(expected_dataset.df.reset_index(drop=True))
        assert all_equal(expected_dataset, filtered_dataset)


class TestCodeFilters:
    def test_python_comment_to_code(self):
        doc_1 = "# Good code\nprint('hello world')"
        doc_2 = "print('bad code')"
        doc_3 = "# Too many\n# comments!"
        doc_4 = "'''Good comment'''\nprint('hello world')"
        dataset = list_to_dataset([doc_1, doc_2, doc_3, doc_4])
        filters = ScoreFilter(PythonCommentToCodeFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_general_commment_to_code(self):
        doc_1 = '// Good code\nprintf("hello world\\n")'
        doc_2 = 'printf("bad code\\n")'
        doc_3 = "// Way far too many\n// comments!"
        doc_4 = '/*\nGood comment\n*/\nprintf("hello world\\n")'
        dataset = list_to_dataset([doc_1, doc_2, doc_3, doc_4])
        filters = ScoreFilter(GeneralCommentToCodeFilter("text/x-c++"))
        filtered_data = filters(dataset)

        expected_indices = [0, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_number_lines_code(self):
        doc_1 = """print("too short")"""
        doc_2 = """print("just")
        print("right")"""
        doc_3 = """print("way")
        print("too")
        print("long")
        print("!")"""
        dataset = list_to_dataset([doc_1, doc_2, doc_3])
        filters = ScoreFilter(NumberOfLinesOfCodeFilter(min_lines=2, max_lines=3))
        filtered_data = filters(dataset)

        expected_indices = [1]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_xml_header(self):
        dataset = list_to_dataset(
            ["no header", "<?xml version=1.0>", "slightly offset <?xml version="]
        )
        filters = ScoreFilter(XMLHeaderFilter())
        filtered_data = filters(dataset)

        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_alpha(self):
        dataset = list_to_dataset(["full of alphabet", "<>?$#@!", "mixed <>"])
        filters = ScoreFilter(AlphaFilter())
        filtered_data = filters(dataset)

        expected_indices = [0, 2]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_html_boilerplate(self):
        good_doc = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>Sample Webpage</title>
        </head>
        <body>
            <h1>Welcome to my sample webpage</h1>
            <p>This is a very fun paragraph on my sample webpage.</p>
        </body>
        </html>
        """
        boilerplate_heavy_doc = """
        <!DOCTYPE html>
        <html>
        <head>
            <title>Boilerplate Webpage</title>
        </head>
        <body>
            <h1><span>Welcome</span> <span>to</span> <span>my</span> <span>boilerplate</span> <span>webpage</span></h1>
            <div>
                <div>
                    <div><p>hi</p></div>
                </div>
                <div>
                    <div><p>hi</p></div>
                </div>
            </div>
        </body>
        </html>
        """
        small_doc = """
            <!DOCTYPE html>
            <html><body>hello world</body></html>
        """
        dataset = list_to_dataset([good_doc, boilerplate_heavy_doc, small_doc])
        filters = ScoreFilter(HTMLBoilerplateFilter())
        filtered_data = filters(dataset)

        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    @pytest.fixture
    def per_extension_filter(self):
        metadata_file = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "nemo_curator",
                "utils",
                "code_meta.csv",
            )
        )

        return PerExtensionFilter("c++", "cpp", metadata_file=metadata_file)

    def test_per_extension_filter(self, per_extension_filter):
        good_cpp = """
        #include <iostream>
        using namespace std;
        int main() {
            cout << "Hello World!" << endl;
            return 0;
        };
        """
        dataset = list_to_dataset([good_cpp])
        filters = ScoreFilter(per_extension_filter)
        filtered_data = filters(dataset)
        expected_indices = [0]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])

        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    @pytest.mark.parametrize(
        "content,expected",
        [
            ("", (0, 0.0)),
            ("\n", (0, 0.0)),
            ("abc\n", (3, 1.5)),
            ("Lorem ipsum \ndolor sit amet,", (15, 13.5)),
        ],
    )
    def test_line_statistics(self, per_extension_filter, content, expected):
        line_statistics = per_extension_filter._line_statistics(content)
        assert (
            line_statistics == expected
        ), f"Expected {expected} but got {line_statistics}"


class FakeQualityFilter(DocumentFilter):
    """
    Emulates FastTextQualityFilter without a model
    """

    def __init__(self, alpha=3, seed=42):
        super().__init__()
        self._alpha = alpha
        self._seed = np.random.seed(seed)

    @batched
    def score_document(self, df):
        return pd.Series(np.arange(len(df)) / len(df))

    @batched
    def keep_document(self, df):
        return np.random.pareto(self._alpha, size=len(df)) > 1 - df


class FakeLangId(DocumentFilter):
    """
    Emulates FastTextLangId without a model
    """

    def __init__(self, min_langid_score=0.3, convert_string=False):
        super().__init__()
        self._cutoff = min_langid_score

        # Dask will automatically convert the list score type
        # to a string without this option.
        # See https://github.com/NVIDIA/NeMo-Curator/issues/33
        dask.config.set({"dataframe.convert-string": convert_string})

    @batched
    def score_document(self, df):
        scores = [[0.5, "EN"], [0.7, "HI"], [0.2, "PT"]]
        scores = scores * len(df)
        scores = scores[: len(df)]
        return pd.Series(scores)

    def keep_document(self, score):
        return score[0] >= self._cutoff


class TestClassifierFilters:
    def test_fake_quality_filter(self):
        dataset = list_to_dataset(["a", "b", "c", "d"], npartitions=1)
        filters = ScoreFilter(FakeQualityFilter())
        filtered_data = filters(dataset)

        expected_indices = [1, 2, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    def test_fake_langid_filter(self):
        dataset = list_to_dataset(["a", "b", "c", "d"], npartitions=1)
        filters = ScoreFilter(FakeLangId())
        filtered_data = filters(dataset)

        expected_indices = [0, 1, 3]
        expected_data = DocumentDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"

    @pytest.mark.skipif(
        is_unavailable(comet), reason="Test depends on COMET but it's not installed."
    )
    def test_comet_qe_filter(self):
        dataset = two_lists_to_parallel_dataset(
            [
                "This sentence will be translated on the Chinese side.",
                "This sentence will have something irrelevant on the Chinese side.",
            ],
            [
                "这句话在中文一侧会被翻译。",
                "至尊戒，驭众戒；至尊戒，寻众戒；魔戒至尊引众戒，禁锢众戒黑暗中。",
            ],
            "en",
            "zh",
        )

        from nemo_curator.filters import QualityEstimationFilter
        from nemo_curator.utils.distributed_utils import get_client

        client = get_client(n_workers=1)
        filter_ = QualityEstimationFilter(
            "comet-qe",
            cutoff=-0.25,
            mode="bidi",
            score_type=float,
            metadata_fields=["src_lang", "tgt_lang"],
        )
        filtered_data = filter_(dataset)

        expected_indices = [0]
        expected_data = ParallelDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"
        client.close()

    @pytest.mark.skipif(
        is_unavailable(pymarian),
        reason="Test depends on PyMarian but it's not installed.",
    )
    def test_cometoid_qe_filter(self):
        dataset = two_lists_to_parallel_dataset(
            [
                "This sentence will be translated on the Chinese side.",
                "This sentence will have something irrelevant on the Chinese side.",
            ],
            [
                "这句话在中文一侧会被翻译。",
                "至尊戒，驭众戒；至尊戒，寻众戒；魔戒至尊引众戒，禁锢众戒黑暗中。",
            ],
            "en",
            "zh",
        )

        from nemo_curator.filters import QualityEstimationFilter
        from nemo_curator.utils.distributed_utils import get_client

        client = get_client(n_workers=1)
        filter_ = QualityEstimationFilter(
            "cometoid-wmt23",
            cutoff=0.75,
            mode="bidi",
            score_type=float,
            metadata_fields=["src_lang", "tgt_lang"],
        )  # enable GPU by gpu=True
        filtered_data = filter_(dataset)

        expected_indices = [0]
        expected_data = ParallelDataset(dataset.df.loc[expected_indices])
        assert all_equal(
            expected_data, filtered_data
        ), f"Expected {expected_data} but got {filtered_data}"
        client.close()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from typing import Callable, Union

import cudf
import cupy as cp
import torch
from crossfit.backend.cudf.series import create_list_series_from_1d_or_2d_ar

from nemo_curator.datasets import ImageTextPairDataset
from nemo_curator.utils.distributed_utils import load_object_on_worker


class ImageClassifier(ABC):
    """
    An abstract base class that represents a classifier on top
    of embeddings generated by a CLIP vision encoder.

    Subclasses only need to define how a model is loaded.
    They may also override the postprocess method if they would like
    to modify output series of predictions before it gets combined into
    the dataset. The classifier must be able to fit on a single GPU.
    """

    def __init__(
        self,
        model_name: str,
        embedding_column: str,
        pred_column: str,
        pred_type: Union[str, type],
        batch_size: int,
        embedding_size: int,
    ) -> None:
        """
        Constructs an image classifier.

        Args:
            model_name (str): A unqiue name to identify the model on each worker
                and in the logs.
            embedding_column (str): The column name that stores the image
                embeddings.
            pred_column (str): The column name to be added where the classifier's
                predictions will be stored.
            pred_type (Union[str, type]): The datatype of the pred_column.
            batch_size (int): If greater than 0, the image embeddings
                will be processed in batches of at most this size. If less than 0,
                all embeddings will be processed at once.
        """
        self.model_name = model_name
        self.embedding_column = embedding_column
        self.pred_column = pred_column
        self.pred_type = pred_type
        self.batch_size = batch_size
        self.embedding_size = embedding_size

    def __call__(self, dataset: ImageTextPairDataset) -> ImageTextPairDataset:
        """
        Classifies all embeddings in the dataset.

        Args:
            dataset (ImageTextPairDataset): The dataset to classify.

        Returns:
            ImageTextPairDataset: A dataset with classifier scores.
        """
        meta = dataset.metadata.dtypes.to_dict()
        meta[self.pred_column] = self.pred_type
        embedding_df = dataset.metadata.map_partitions(self._run_inference, meta=meta)

        return ImageTextPairDataset(
            dataset.path,
            metadata=embedding_df,
            tar_files=dataset.tar_files,
            id_col=dataset.id_col,
        )

    def _run_inference(self, partition, partition_info=None):
        device = "cuda"

        model = load_object_on_worker(
            self.model_name,
            self.load_model,
            {"device": device},
        )

        embeddings = torch.as_tensor(
            partition[self.embedding_column].list.leaves.values.reshape(
                len(partition), -1
            ),
            device=device,
        )

        if self.embedding_size != embeddings.shape[-1]:
            raise RuntimeError(
                f"{self.model_name} expects embedding size {self.embedding_size} but column "
                f"'{self.embedding_column}' has embedding size {embeddings.shape[-1]}. Ensure your "
                "classifier is compatible with the CLIP model you used to generate the embeddings."
            )

        with torch.no_grad():
            if self.batch_size > 0:
                batches = torch.split(embeddings, self.batch_size)
                model_results = []
                for batch in batches:
                    batch_results = model(batch)
                    model_results.append(batch_results)
                scores = torch.cat(model_results, dim=0)
            else:
                scores = model(embeddings)

        scores = cp.asarray(scores)

        series = create_list_series_from_1d_or_2d_ar(scores, index=partition.index)
        postprocessed_series = self.postprocess(series)
        partition[self.pred_column] = postprocessed_series

        print(
            f"Partition {partition_info['number']} - Classification with {self.model_name} completed for {len(scores)} samples."
        )

        return partition

    @abstractmethod
    def load_model(self, device: str) -> Callable:
        """
        Loads the classifier model.

        Args:
            device (str): A PyTorch device identifier that specifies what GPU
                to load the model on.

        Returns:
            Callable: A callable model, usually a torch.nn.Module.
                The input to this model will be the batches of images output
                by the ImageEmbedder.load_dataset_shard.
        """
        pass

    def postprocess(self, series: cudf.Series) -> cudf.Series:
        """
        Postprocesses the predictions of the classifier before saving
        them to the metadata.

        Args:
            series (cudf.Series): The cuDF series of raw model predictions.

        Returns:
            cudf.Series: The same series unmodified. Override in your classifier
                if needed.
        """
        return series



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
import warnings
from typing import Dict

import requests
from bs4 import BeautifulSoup, MarkupResemblesLocatorWarning

from nemo_curator.download.doc_builder import (
    DocumentDownloader,
    DocumentExtractor,
    DocumentIterator,
)

# Ignore the specific BeautifulSoup warning
warnings.filterwarnings("ignore", category=MarkupResemblesLocatorWarning)


class LawQADownloader(DocumentDownloader):
    """
    A class for downloading Law QA dataset.
    """

    def __init__(self, download_dir: str):
        super().__init__()

        if not os.path.isdir(download_dir):
            os.makedirs(download_dir)

        self._download_dir = download_dir
        print("Download directory: ", self._download_dir)

    def download(self, url: str) -> str:
        """
        Downloads the Law QA dataset from the given URL.

        Args:
            url (str): The URL of the Law QA dataset.

        Returns:
            str: The path of the downloaded file.

        """
        filename = os.path.basename(url)
        output_file = os.path.join(self._download_dir, filename)

        if os.path.exists(output_file):
            print(f"File '{output_file}' already exists, skipping download.")
            return output_file

        print(f"Downloading Law QA dataset from '{url}'...")
        response = requests.get(url)

        with open(output_file, "wb") as file:
            file.write(response.content)

        return output_file


class LawQAIterator(DocumentIterator):

    def __init__(self):
        super().__init__()
        self._counter = -1
        self._extractor = LawQAExtractor()

    def iterate(self, file_path):
        """
        Iterates over the content of a file and yields extracted records.

        Args:
            file_path (str): The path to the file to be iterated.

        Yields:
            dict: A dictionary representing a record extracted from the file.
        """
        self._counter = -1
        file_name = os.path.basename(file_path)

        with open(file_path, "r", encoding="utf-8") as file:
            lines = file.readlines()

        file_content = "".join(lines)
        json_content = json.loads(file_content)

        for row in json_content:
            self._counter += 1
            extracted_content = self._extractor.extract(row)

            # Skip if the question has no answers.
            if extracted_content is None:
                continue

            id, extracted_content = extracted_content
            meta = {
                "file_name": file_name,
                "id": f"law-stackexchange-qa-{id}",
            }

            record = {**meta, **extracted_content}
            yield record


class LawQAExtractor(DocumentExtractor):

    def extract(self, content: str) -> Dict[str, str]:
        """
        Extracts relevant information from a law-related question and its best answer.

        Args:
            content (str): The content of the question and its answers.

        Returns:
            Dict[str, str]: A dictionary containing the extracted information, including the question ID, title, body,
            score, best answer, best answer score, and tags.
        """
        id = content["question_id"]
        q_title = content["question_title"]
        q_body = content["question_body"]
        q_score = content["score"]
        tags = ",".join(sorted(content["tags"]))

        # If this question has no answers, skip it.
        if len(content["answers"]) == 0:
            return None

        # All answers are sorted by votes, so take the first answer as the best one.
        best_answer = content["answers"][0]
        best_answer_score = best_answer["score"]
        best_answer = best_answer["body"]

        # Get rid of HTML tags using beautifulsoup
        # NOTE: Doing this here so that I can split the dataset without having to worry about curating the test split.
        q_title = self._clean_html(q_title)
        q_body = self._clean_html(q_body)
        best_answer = self._clean_html(best_answer)

        return id, {
            "title": q_title,
            "question": q_body,
            "question_score": q_score,
            "answer": best_answer,
            "answer_score": best_answer_score,
            "tags": tags,
        }

    def _clean_html(self, text: str) -> str:
        text = BeautifulSoup(text, "lxml").get_text()
        return re.sub(r"\s+", " ", text).strip()



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC, abstractmethod
from typing import Iterable, List, Optional, Union

from nemo_curator.services.conversation_formatter import ConversationFormatter


class LLMClient(ABC):
    """
    Interface representing a client connecting to an LLM inference server
    and making requests synchronously
    """

    @abstractmethod
    def query_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
        max_tokens: Optional[int] = None,
        n: Optional[int] = 1,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
    ) -> List[str]:
        raise NotImplementedError("Subclass of LLMClient must implement 'query_model'")

    @abstractmethod
    def query_reward_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
    ) -> dict:
        raise NotImplementedError(
            "Subclass of LLMClient must implement 'query_reward_model'"
        )


class AsyncLLMClient(ABC):
    """
    Interface representing a client connecting to an LLM inference server
    and making requests asynchronously
    """

    @abstractmethod
    async def query_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
        max_tokens: Optional[int] = None,
        n: Optional[int] = 1,
        seed: Optional[int] = None,
        stop: Union[Optional[str], List[str]] = None,
        stream: bool = False,
        temperature: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
    ) -> List[str]:
        raise NotImplementedError(
            "Subclass of AsyncLLMClient must implement 'query_model'"
        )

    @abstractmethod
    async def query_reward_model(
        self,
        *,
        messages: Iterable,
        model: str,
        conversation_formatter: Optional[ConversationFormatter] = None,
    ) -> dict:
        raise NotImplementedError(
            "Subclass of LLMClient must implement 'query_reward_model'"
        )



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

import pytest
import yaml

from nemo_curator.modules.config import BaseConfig


@dataclass
class CustomConfig(BaseConfig):
    a: str
    b: int
    c: bool
    d: float = 3.0

    def __post_init__(self):
        if self.d <= 0:
            raise ValueError("d must be positive")


class TestConfig:
    @pytest.fixture(autouse=True)
    def config_params(self):
        self.config_dict = {"a": "a", "b": 1, "c": True, "d": 4.0}

    def test_init(self):
        config = CustomConfig(a="a", b=1, c=True)
        assert config.a == "a"
        assert config.b == 1
        assert config.c is True
        assert config.d == 3.0

    def test_from_yaml(self, tmpdir):
        with open(tmpdir / "test_config.yaml", "w") as file:
            yaml.dump(self.config_dict, file)

        config = CustomConfig.from_yaml(tmpdir / "test_config.yaml")
        for key, value in self.config_dict.items():
            assert getattr(config, key) == value

    def test_from_yaml_raises(self, tmpdir):
        config_dict = self.config_dict.copy()
        config_dict["d"] = -1.0
        with open(tmpdir / "test_config.yaml", "w") as file:
            yaml.dump(config_dict, file)
        with pytest.raises(ValueError):
            CustomConfig.from_yaml(tmpdir / "test_config.yaml")

    def test_from_yaml_missing_key(self, tmpdir):
        config_dict = self.config_dict.copy()
        del config_dict["a"]
        with open(tmpdir / "test_config.yaml", "w") as file:
            yaml.dump(config_dict, file)
        with pytest.raises(TypeError):
            CustomConfig.from_yaml(tmpdir / "test_config.yaml")

    def test_from_yaml_extra_key(self, tmpdir):
        config_dict = self.config_dict.copy()
        config_dict["e"] = "e"
        with open(tmpdir / "test_config.yaml", "w") as file:
            yaml.dump(config_dict, file)
        with pytest.raises(TypeError):
            CustomConfig.from_yaml(tmpdir / "test_config.yaml")

    def test_post_init_raises(self):
        with pytest.raises(ValueError):
            CustomConfig(a="a", b=1, c=True, d=-1.0)



import ctypes
import gc
import logging
import os
from pathlib import Path

from dask.distributed import Client, LocalCluster
from helper import process_data

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)

DATA_BASE = os.environ.get("DATA_BASE")
INPUT_BASE = os.path.join(DATA_BASE, "raw/fineweb-edu-score-2/data")
OUTPUT_BASE = os.path.join(DATA_BASE, "processed/fineweb-edu-score-2")
CPU_WORKERS = os.environ.get("CPU_WORKERS")


def trim_memory() -> int:
    libc = ctypes.CDLL("libc.so.6")
    return libc.malloc_trim(0)


def get_folder_size(folder_path):
    return sum(
        file.stat().st_size for file in Path(folder_path).rglob("*") if file.is_file()
    )


def sort_folders_by_size(parent_directory):
    folders = [
        f
        for f in os.listdir(parent_directory)
        if os.path.isdir(os.path.join(parent_directory, f))
    ]
    folder_sizes = [
        (folder, get_folder_size(os.path.join(parent_directory, folder)))
        for folder in folders
    ]
    return sorted(folder_sizes, key=lambda x: x[1])


def bytes_to_human_readable(size_in_bytes):
    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
    suffix_index = 0
    size = float(size_in_bytes)
    while size >= 1024 and suffix_index < len(suffixes) - 1:
        size /= 1024.0
        suffix_index += 1
    return f"{size:.2f} {suffixes[suffix_index]}"


if __name__ == "__main__":
    logging.info("Starting Dask cluster")
    cluster = LocalCluster(n_workers=CPU_WORKERS, processes=True, memory_limit="240GB")
    client = Client(cluster)
    logging.info(client)

    components_with_sizes = sort_folders_by_size(INPUT_BASE)

    for component, component_size in components_with_sizes:
        input_path = os.path.join(INPUT_BASE, component)
        if not os.path.exists(input_path) or not os.path.isdir(input_path):
            continue
        output_path = os.path.join(OUTPUT_BASE, component)
        logging.info(
            f"Processing {component}, size = {bytes_to_human_readable(component_size)}"
        )
        process_data(
            input_folder=input_path,
            output_folder=output_path,
            prefix=f"fwe2-{component}",
        )
        logging.info("Trimming memory")
        gc.collect()
        client.run(trim_memory)
        logging.info("Done!")

    client.cluster.close()
    client.shutdown()



import logging
import os
import time

import dask.dataframe as dd
import pyarrow as pa
from dask.distributed import Client, LocalCluster

logging.basicConfig(format="%(asctime)s: %(message)s", level=logging.INFO)

DATA_BASE = os.environ.get("DATA_BASE")
CPU_WORKERS = os.environ.get("CPU_WORKERS")


if __name__ == "__main__":
    t0 = time.time()
    cluster = LocalCluster(n_workers=CPU_WORKERS, threads_per_worker=2, processes=True)
    client = Client(cluster)

    logging.info(f"Filtering...")
    INPUT_BASE = os.path.join(DATA_BASE, "processed/fineweb-edu-score-2")
    OUTPUT_BASE = os.path.join(DATA_BASE, "zyda2/fwe3")
    folders = sorted(os.listdir(INPUT_BASE))
    for folder in folders:
        print(f"\nProcessing {folder}")
        ddf = dd.read_parquet(os.path.join(INPUT_BASE, folder))
        ddf_filtered = ddf[ddf["int_score"] >= 3].repartition(partition_size="512M")
        out_folder = os.path.join(OUTPUT_BASE, folder)
        print(f"Saving to {out_folder}")
        ddf_filtered.to_parquet(out_folder, write_index=False, overwrite=True)
    logging.info(f"Done in {time.time() - t0:.2f} sec")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from nemo_curator.modifiers.async_llm_pii_modifier import AsyncLLMPiiModifier
from nemo_curator.modifiers.llm_pii_modifier import LLMPiiModifier
from nemo_curator.utils.llm_pii_utils import (
    EntitySpan,
    find_entity_spans,
    fix_overlaps,
    get_system_prompt,
    redact,
    validate_entity,
    validate_keys,
)


class TestLLMPiiUtils:
    def test_validate_keys(self):
        valid_keys = {"entity_type": "date_time", "entity_text": "1989.12.22"}
        assert validate_keys(valid_keys)

        incomplete_keys = {"entity_type": "date_time"}
        assert not validate_keys(incomplete_keys)

        invalid_keys = {"type": "date_time", "text": "1989.12.22"}
        assert not validate_keys(invalid_keys)

    def test_validate_entity(self):
        valid_entity = {"entity_type": "date_time", "entity_text": "1989.12.22"}
        text = "Unloading Plan for Shipment MRN-293104, MED25315002, dated 1989.12.22. Driver EMP730359, Vehicle KS40540825."
        min_length = 2
        assert validate_entity(valid_entity, text, min_length)

        invalid_entity = {"type": "date_time", "text": "1989.12.22"}
        assert not validate_entity(invalid_entity, text, min_length)

        invalid_entity_text = {"entity_type": "date_time", "entity_text": "1999.12.22"}
        assert not validate_entity(invalid_entity_text, text, min_length)

        short_entity_text = {"entity_type": "date_time", "entity_text": "1"}
        assert not validate_entity(short_entity_text, text, min_length)

    def test_redact(self):
        full_text = "Transaction details: gasLimit set to 1000000 units by tw_brian740, gasPrice set to 10 Gwei by veronicawood@example.org, contactable at +1-869-341-9301x7005, located at Suite 378, Yolanda Mountain, Burkeberg."
        pii_entities = [
            {"entity_type": "name", "entity_text": "tw_brian740"},
            {"entity_type": "name", "entity_text": "veronicawood"},
            {"entity_type": "email", "entity_text": "veronicawood@example.org"},
            {"entity_type": "phone_number", "entity_text": "+1-869-341-9301x7005"},
            {
                "entity_type": "location",
                "entity_text": "Suite 378, Yolanda Mountain, Burkeberg",
            },
        ]
        redacted_text = redact(full_text, pii_entities)
        expected_text = "Transaction details: gasLimit set to 1000000 units by {{name}}, gasPrice set to 10 Gwei by {{email}}, contactable at {{phone_number}}, located at {{location}}."
        assert redacted_text == expected_text

    def test_find_entity_spans(self):
        text = "Transaction details: gasLimit set to 1000000 units by tw_brian740, gasPrice set to 10 Gwei by veronicawood@example.org, contactable at +1-869-341-9301x7005, located at Suite 378, Yolanda Mountain, Burkeberg."
        entities = [
            {"entity_type": "name", "entity_text": "tw_brian740"},
            # Add repeated entity to check robustness
            {"entity_type": "name", "entity_text": "tw_brian740"},
            {"entity_type": "name", "entity_text": "veronicawood"},
            {"entity_type": "email", "entity_text": "veronicawood@example.org"},
            {"entity_type": "phone_number", "entity_text": "+1-869-341-9301x7005"},
            {
                "entity_type": "location",
                "entity_text": "Suite 378, Yolanda Mountain, Burkeberg",
            },
        ]
        spans = find_entity_spans(text, entities)
        assert spans == [
            EntitySpan(entity_type="name", start_position=54, end_position=65),
            EntitySpan(entity_type="name", start_position=94, end_position=106),
            EntitySpan(entity_type="email", start_position=94, end_position=118),
            EntitySpan(
                entity_type="phone_number", start_position=135, end_position=155
            ),
            EntitySpan(entity_type="location", start_position=168, end_position=206),
        ]

    def test_fix_overlaps(self):
        assert fix_overlaps([]) == []

        spans = [
            EntitySpan(entity_type="name", start_position=54, end_position=65),
            EntitySpan(entity_type="name", start_position=94, end_position=106),
            EntitySpan(entity_type="email", start_position=94, end_position=118),
            EntitySpan(
                entity_type="phone_number", start_position=135, end_position=155
            ),
            EntitySpan(entity_type="location", start_position=168, end_position=206),
        ]
        fixed_spans = fix_overlaps(spans)
        assert fixed_spans == [
            EntitySpan(entity_type="name", start_position=54, end_position=65),
            EntitySpan(entity_type="email", start_position=94, end_position=118),
            EntitySpan(
                entity_type="phone_number", start_position=135, end_position=155
            ),
            EntitySpan(entity_type="location", start_position=168, end_position=206),
        ]

        spans = [
            EntitySpan(entity_type="date_time", start_position=59, end_position=69),
            EntitySpan(entity_type="employee_id", start_position=78, end_position=87),
            EntitySpan(
                entity_type="medical_record_number", start_position=28, end_position=38
            ),
            EntitySpan(
                entity_type="vehicle_identifier", start_position=97, end_position=107
            ),
        ]
        fixed_spans = fix_overlaps(spans)
        assert fixed_spans == [
            EntitySpan(
                entity_type="medical_record_number", start_position=28, end_position=38
            ),
            EntitySpan(entity_type="date_time", start_position=59, end_position=69),
            EntitySpan(entity_type="employee_id", start_position=78, end_position=87),
            EntitySpan(
                entity_type="vehicle_identifier", start_position=97, end_position=107
            ),
        ]


class TestLLMPiiModifier:
    def test_system_prompt(self):
        # Default system prompt, default PII labels, and English
        modifier = LLMPiiModifier(base_url="https://integrate.api.nvidia.com/v1")
        assert modifier.system_prompt == get_system_prompt()

        # Default system prompt, default PII labels, and non-English
        with pytest.warns(
            UserWarning,
            match="The default system prompt is only available for English text",
        ):
            modifier = LLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1", language="fr"
            )
            assert modifier.system_prompt == get_system_prompt()

        # Custom system prompt, default PII labels, and English
        with pytest.warns(
            UserWarning,
            match="Using the default system prompt is strongly recommended for English text",
        ):
            system_prompt = "You are a helpful assistant."
            modifier = LLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1",
                system_prompt=system_prompt,
            )
            assert modifier.system_prompt == system_prompt

        # Custom system prompt, default PII labels, and non-English
        system_prompt = "Vous êtes un assistant utile."
        modifier = LLMPiiModifier(
            base_url="https://integrate.api.nvidia.com/v1", system_prompt=system_prompt
        )
        assert modifier.system_prompt == system_prompt

        # Default system prompt and custom PII labels
        pii_labels = ["name", "email", "phone_number", "location"]
        modifier = LLMPiiModifier(
            base_url="https://integrate.api.nvidia.com/v1", pii_labels=pii_labels
        )
        assert modifier.system_prompt == get_system_prompt(pii_labels)

        # Custom system prompt and custom PII labels
        with pytest.warns(
            UserWarning,
            match="Custom system_prompt and custom pii_labels were both provided",
        ):
            system_prompt = "You are a helpful assistant."
            pii_labels = ["name", "email", "phone_number", "location"]
            modifier = LLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1",
                system_prompt=system_prompt,
                pii_labels=pii_labels,
            )
            assert modifier.system_prompt == system_prompt


class TestAsyncLLMPiiModifier:
    def test_system_prompt(self):
        # Default system prompt, default PII labels, and English
        modifier = AsyncLLMPiiModifier(base_url="https://integrate.api.nvidia.com/v1")
        assert modifier.system_prompt == get_system_prompt()

        # Default system prompt, default PII labels, and non-English
        with pytest.warns(
            UserWarning,
            match="The default system prompt is only available for English text",
        ):
            modifier = AsyncLLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1", language="fr"
            )
            assert modifier.system_prompt == get_system_prompt()

        # Custom system prompt, default PII labels, and English
        with pytest.warns(
            UserWarning,
            match="Using the default system prompt is strongly recommended for English text",
        ):
            system_prompt = "You are a helpful assistant."
            modifier = AsyncLLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1",
                system_prompt=system_prompt,
            )
            assert modifier.system_prompt == system_prompt

        # Custom system prompt, default PII labels, and non-English
        system_prompt = "Vous êtes un assistant utile."
        modifier = AsyncLLMPiiModifier(
            base_url="https://integrate.api.nvidia.com/v1", system_prompt=system_prompt
        )
        assert modifier.system_prompt == system_prompt

        # Default system prompt and custom PII labels
        pii_labels = ["name", "email", "phone_number", "location"]
        modifier = AsyncLLMPiiModifier(
            base_url="https://integrate.api.nvidia.com/v1", pii_labels=pii_labels
        )
        assert modifier.system_prompt == get_system_prompt(pii_labels)

        # Custom system prompt and custom PII labels
        with pytest.warns(
            UserWarning,
            match="Custom system_prompt and custom pii_labels were both provided",
        ):
            system_prompt = "You are a helpful assistant."
            pii_labels = ["name", "email", "phone_number", "location"]
            modifier = AsyncLLMPiiModifier(
                base_url="https://integrate.api.nvidia.com/v1",
                system_prompt=system_prompt,
                pii_labels=pii_labels,
            )
            assert modifier.system_prompt == system_prompt



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import random
from typing import TYPE_CHECKING, Literal

import dask
import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest
import torch
import torch.nn.functional as F
from dask.dataframe.utils import assert_eq
from sklearn.datasets import make_blobs
from transformers import AutoConfig, AutoModel, AutoTokenizer

from nemo_curator import SemDedup, SemDedupConfig
from nemo_curator.datasets import DocumentDataset
from nemo_curator.utils.import_utils import gpu_only_import, gpu_only_import_from

cudf = gpu_only_import("cudf")
dask_cudf = gpu_only_import("dask_cudf")
cp = gpu_only_import("cupy")
EmbeddingCreator = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.embeddings", "EmbeddingCreator"
)
ClusteringModel = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.clusteringmodel", "ClusteringModel"
)
SemanticClusterLevelDedup = gpu_only_import_from(
    "nemo_curator.modules.semantic_dedup.semanticclusterleveldedup",
    "SemanticClusterLevelDedup",
)
add_l2_cosine_dist_to_centroid = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils",
    "add_l2_cosine_dist_to_centroid",
)
normalize_embeddings_col_in_df = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils", "normalize_embeddings_col_in_df"
)
get_array_from_df = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils", "get_array_from_df"
)
pairwise_cosine_similarity = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils", "pairwise_cosine_similarity"
)
pairwise_cosine_similarity_batched = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils", "pairwise_cosine_similarity_batched"
)
get_semantic_matches_per_cluster = gpu_only_import_from(
    "nemo_curator.utils.semdedup_utils", "get_semantic_matches_per_cluster"
)

if TYPE_CHECKING:
    import cupy as cp

    from nemo_curator.modules.semantic_dedup.clusteringmodel import ClusteringModel
    from nemo_curator.modules.semantic_dedup.semanticclusterleveldedup import (
        SemanticClusterLevelDedup,
    )
    from nemo_curator.modules.semantic_dedup.semdedup import SemDedup
    from nemo_curator.utils.semdedup_utils import (
        add_l2_cosine_dist_to_centroid,
        get_array_from_df,
        get_semantic_matches_per_cluster,
        normalize_embeddings_col_in_df,
        pairwise_cosine_similarity,
        pairwise_cosine_similarity_batched,
    )


@pytest.fixture
def dedup_data():
    df = cudf.DataFrame(
        {
            "id": [1, 2, 3, 4, 100, 200, 300],
            "text": [
                "The quick brown fox jumps over the lazy dog",
                "The quick brown foxes jumps over the lazy dog",
                "The quick brown wolf jumps over the lazy dog",
                "The quick black cat jumps over the lazy dog",
                "A test string",
                "Another test string",
                "A different object",
            ],
        }
    )
    df = dask_cudf.from_cudf(df, 2)
    return DocumentDataset(df)


@pytest.fixture
def non_dedup_data():
    df = cudf.DataFrame(
        {
            "doc_id": ["doc_1", "doc_2"],
            "text": [
                "The quick brown fox jumps over the lazy dog",
                "A test string",
            ],
        }
    )
    df = dask_cudf.from_cudf(df, 2)
    return DocumentDataset(df)


@pytest.mark.gpu
class TestSemDuplicates:
    @pytest.mark.parametrize("n_clusters", [3, 10])
    @pytest.mark.parametrize("id_col_type", ["int", "str"])
    @pytest.mark.parametrize("perform_removal", [True, False])
    def test_sem_dedup(
        self,
        dedup_data,
        tmpdir,
        n_clusters,
        id_col_type,
        perform_removal,
        gpu_client,
    ):
        cache_dir = os.path.join(tmpdir, "test_sem_dedup_cache")
        config = SemDedupConfig(
            cache_dir=cache_dir,
            n_clusters=n_clusters,
            eps_to_extract=0.10,
        )

        sem_duplicates = SemDedup(
            config=config,
            input_column="text",
            id_column="id",
            perform_removal=perform_removal,
        )
        # Convert id column to the specified type
        dedup_data.df["id"] = dedup_data.df["id"].astype(id_col_type)

        dedup_data_len = dedup_data.df.shape[0].compute()
        if n_clusters > dedup_data_len:
            # Number of records in the dataset should never be less than n_clusters
            with pytest.raises(ValueError):
                result = sem_duplicates(dedup_data)
        else:
            # Correctly returns the original dataset with no duplicates removed
            result = sem_duplicates(dedup_data)
            result_df = result.df.compute()
            docs_to_remove = [1, 100]
            if id_col_type == "str":
                docs_to_remove = list(map(str, docs_to_remove))

            if not perform_removal:
                expected_df = cudf.Series(docs_to_remove, name="id", dtype=id_col_type)
                assert_eq(result_df["id"].sort_values(), expected_df, check_index=False)
            else:
                assert_eq(
                    result_df,
                    dedup_data.df[~dedup_data.df["id"].isin(docs_to_remove)],
                    check_index=False,
                )

    @pytest.mark.parametrize("n_clusters", [2, 3])
    @pytest.mark.parametrize("perform_removal", [True, False])
    def test_no_sem_dedup(
        self,
        non_dedup_data,
        tmpdir,
        n_clusters,
        perform_removal,
        gpu_client,
    ):
        cache_dir = os.path.join(tmpdir, "test_no_sem_dedup")
        config = SemDedupConfig(
            cache_dir=cache_dir,
            n_clusters=n_clusters,
            eps_to_extract=0.10,
        )

        sem_duplicates = SemDedup(
            config=config,
            input_column="text",
            id_column="doc_id",
            perform_removal=perform_removal,
        )

        non_dedup_data_len = non_dedup_data.df.shape[0].compute()
        if n_clusters > non_dedup_data_len:
            # Number of records in the dataset should never be less than n_clusters
            with pytest.raises(ValueError):
                result = sem_duplicates(non_dedup_data)
        else:
            # Correctly returns the original dataset with no duplicates removed
            result = sem_duplicates(non_dedup_data)
            result_df = result.df.compute()
            if not perform_removal:
                assert len(result_df) == 0
            else:
                assert_eq(result_df, non_dedup_data.df, check_index=False)

    @pytest.mark.parametrize("pooling_strategy", ["last_token", "mean_pooling"])
    def test_embedding_creator_pooling_strategies(self, tmpdir, pooling_strategy):
        test_text_1 = "The quick brown fox jumps over the lazy dog"
        test_text_2 = "The brown fox jumps over the dog"
        test_texts = [test_text_1, test_text_2] * 32
        df = cudf.DataFrame({"text": test_texts})
        ddf = dask_cudf.from_cudf(df, 1)

        cache_dir = os.path.join(tmpdir, "test_embeddings_cache")

        embedding_creator = EmbeddingCreator(
            embedding_model_name_or_path="sentence-transformers/all-MiniLM-L6-v2",
            embedding_batch_size=32,
            embedding_pooling_strategy=pooling_strategy,
            input_column="text",
            embedding_output_dir=os.path.join(cache_dir, "mean_embeddings"),
        )

        embeddings = embedding_creator.create_embeddings(ddf).compute()
        embeddings = embeddings["embeddings"].to_arrow().to_pylist()
        embeddings = np.array(embeddings)

        reference_embeddings = get_reference_embeddings(
            test_texts, pooling_strategy=pooling_strategy
        )

        assert np.allclose(
            embeddings, reference_embeddings, atol=1e-3
        ), "Embeddings should match reference embeddings"


def get_reference_embeddings(
    texts,
    model_name="sentence-transformers/all-MiniLM-L6-v2",
    pooling_strategy="last_token",
):
    """
    Get embeddings using either last token or mean pooling strategy.

    Args:
        texts: List of input texts
        model_name: Name or path of the model to use
        pooling_strategy: Either "last_token" for last token or "mean" for mean pooling
    """
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    model = AutoModel.from_pretrained(model_name)
    model = model.to("cuda")
    model.eval()
    max_len_to_use = tokenizer.model_max_length
    if max_len_to_use > 1e5:
        max_len_to_use = AutoConfig.from_pretrained(model_name).max_position_embeddings
    max_seq_length: int = max_len_to_use

    embs = []
    for text in texts:
        inputs = tokenizer(
            text,
            return_tensors="pt",
            padding=True,
            truncation=True,
            max_length=max_seq_length,
        )
        inputs = {k: v.to("cuda") for k, v in inputs.items()}

        with torch.no_grad():
            with torch.autocast(device_type="cuda"):
                outputs = model(**inputs)

        if pooling_strategy == "last_token":
            embeddings = outputs.last_hidden_state[:, -1, :]
        elif pooling_strategy == "mean_pooling":
            token_embeddings = outputs.last_hidden_state
            attention_mask = inputs["attention_mask"]
            input_mask_expanded = (
                attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            )
            sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, dim=1)
            sum_mask = torch.clamp(input_mask_expanded.sum(dim=1), min=1e-9)
            embeddings = sum_embeddings / sum_mask
        else:
            raise ValueError(
                "pooling_strategy must be either 'last_token' or 'mean_pooling'"
            )

        normed_emb = F.normalize(embeddings, dim=1).cpu()
        normed_emb = normed_emb.squeeze(0)
        embs.append(normed_emb)

    return np.array(embs)


@pytest.mark.gpu
class TestSemDedupUtils:
    def setup_method(self):
        # We create a 6x3 array where each row is a unit vector
        # The second and last two rows are the same
        input_embeddings = torch.tensor(
            np.asarray(
                [[1, 2, 3], [4, 5, 6], [7, 8, 9], [10, 11, 12], [1, 2, 3], [1, 2, 3]],
            ),
            dtype=torch.float32,
        )
        # Normalize the input array
        self.input_embeddings = input_embeddings / torch.norm(
            input_embeddings, dim=1, keepdim=True
        )
        self.expected_pairwise_similarity = np.array(
            [0.0000, 0.974631, 0.998190, 0.999618, 1.0000, 1.0000]
        )
        self.expected_indices = np.array([0, 0, 1, 2, 0, 0])

    def test_pairwise_cosine_similarity(self):
        max_similarity, max_indices = pairwise_cosine_similarity(
            self.input_embeddings, "cuda"
        )
        np.testing.assert_allclose(
            max_similarity.tolist(),
            self.expected_pairwise_similarity,
            rtol=1e-6,
            atol=1e-6,
        )
        np.testing.assert_array_equal(max_indices.tolist(), self.expected_indices)

    @pytest.mark.parametrize("batch_size", [1, 2, 3, 4, 5, 6])
    def test_pairwise_cosine_similarity_batched(self, batch_size: int):
        max_similarity, max_indices = pairwise_cosine_similarity_batched(
            self.input_embeddings, "cuda", batch_size
        )
        np.testing.assert_allclose(
            max_similarity.tolist(),
            self.expected_pairwise_similarity,
            rtol=1e-6,
            atol=1e-6,
        )
        np.testing.assert_array_equal(max_indices.tolist(), self.expected_indices)

    @pytest.mark.parametrize("batch_size", [100, 512, 1024, 2048])
    def test_pairwise_cosine_similarity_batched_rand_array(self, batch_size: int):
        N = 1024
        D = 512
        rand_arr = torch.randn(N, D, device="cuda")
        max_similarity, max_indices = pairwise_cosine_similarity(rand_arr, "cuda")
        max_similarity_batched, max_indices_batched = (
            pairwise_cosine_similarity_batched(rand_arr, "cuda", batch_size=batch_size)
        )
        np.testing.assert_allclose(
            max_similarity.tolist(),
            max_similarity_batched.tolist(),
            rtol=1e-5,
            atol=1e-5,
        )
        np.testing.assert_array_equal(
            max_indices.tolist(), max_indices_batched.tolist()
        )

    def test_get_array_from_df(self):
        df = cudf.DataFrame(
            {
                "embedding": [[3, 4, 5], [1, 2, 2], [1, 0, 0]],
            }
        )
        expected_array = cp.array(
            [
                [3, 4, 5],
                [1, 2, 2],
                [1, 0, 0],
            ]
        )
        cp.testing.assert_allclose(
            get_array_from_df(df, "embedding"), expected_array, rtol=1e-5, atol=1e-5
        )

    def test_normalize_embeddings_col_in_df(self):
        # Mock data setup
        df = cudf.DataFrame(
            {
                "embedding": [[3, 4, 5], [1, 2, 2], [1, 0, 0]],
            }
        )
        expected_normalized = cp.array(
            [
                [0.42426407, 0.565685, 0.707107],
                [0.33333334, 0.6666667, 0.6666667],
                [1.0, 0.0, 0.0],
            ]
        )

        # Call the function
        normalized_embeddings = normalize_embeddings_col_in_df(df, "embedding")

        # Assert the normalized embeddings match the expected values
        cp.testing.assert_allclose(
            get_array_from_df(normalized_embeddings, "embedding"),
            expected_normalized,
            rtol=1e-5,
            atol=1e-5,
        )

    def test_add_l2_cosine_dist_to_centroid(self):
        # Mock data setup
        df = cudf.DataFrame(
            {
                "nearest_cent": [0, 1, 0],
                "embedding": [
                    [1, 0],
                    [0, 1],
                    [0.6, 0.8],
                ],
            }
        )
        centroids = cp.array([[1, 0], [0, 1]])
        # Call the function
        df_with_distances = add_l2_cosine_dist_to_centroid(df, "embedding", centroids)

        # Assert the distances match the expected values
        np.testing.assert_almost_equal(
            df_with_distances["l2_dist_to_cent"].to_arrow().to_pylist(),
            [0.0, 0.0, (0.16 + 0.64) ** 0.5],
            decimal=4,
        )
        np.testing.assert_almost_equal(
            df_with_distances["cosine_dist_to_cent"].to_arrow().to_pylist(),
            [0.0, 0.0, 0.4],
            decimal=4,
        )

    @pytest.mark.parametrize("which_to_keep", ["hard", "easy"])
    @pytest.mark.parametrize("sim_metric", ["cosine", "l2"])
    @pytest.mark.parametrize("id_col_type", ["int", "str"])
    def test_get_semantic_matches_per_cluster(
        self,
        which_to_keep: Literal["hard", "easy"],
        sim_metric: Literal["cosine", "l2"],
        id_col_type: Literal["int", "str"],
        tmpdir,
    ):
        cluster_c = 0
        self.centroid = self.input_embeddings[:1]
        os.makedirs(os.path.join(tmpdir, f"nearest_cent={cluster_c}"), exist_ok=True)
        # Step 1) Simulate rank_within_cluster
        embeddings_df = cudf.DataFrame(
            {
                "embedding": self.input_embeddings.tolist(),
                "id": list(range(self.input_embeddings.shape[0])),
            }
        )
        embeddings_df["id"] = embeddings_df["id"].astype(id_col_type)
        # Step 1) Call add_l2_cosine_dist_to_centroid
        embeddings_df = add_l2_cosine_dist_to_centroid(
            df=embeddings_df.assign(nearest_cent=cluster_c),
            embedding_col="embedding",
            centroids=cp.asarray(self.centroid),
        )
        embeddings_df.to_parquet(
            os.path.join(tmpdir, f"nearest_cent={cluster_c}/file.parquet")
        )
        # Step 2) Call get_semantic_matches_per_cluster
        # this internally calls read_cluster_embeddings_and_sort_by_id and pairwise_cosine_similarity
        get_semantic_matches_per_cluster(
            cluster_id=cluster_c,
            emb_by_clust_dir=tmpdir,
            id_col="id",
            output_dir=tmpdir,
            embedding_col="embedding",
            which_to_keep=which_to_keep,
            sim_metric=sim_metric,
            batched_cosine_similarity=1024,
        )

        # Read the output
        output_df = pd.read_parquet(
            os.path.join(tmpdir, f"cluster_{cluster_c}.parquet")
        )
        # See https://github.com/NVIDIA/NeMo-Curator/issues/610 to understand the expected output and the walkthrough
        if which_to_keep == "hard":
            expected_ids = [3, 2, 1, 5, 4, 0]
            expected_max_ids = [3, 3, 2, 1, 5, 5]
            expected_cosine_sim_scores = np.array(
                [
                    0.0000,
                    0.99961,
                    0.99819,
                    0.974631,
                    1.0000,
                    1.0000,
                ],
                dtype=np.float32,
            )
        else:
            expected_ids = [0, 4, 5, 1, 2, 3]
            expected_max_ids = [0, 0, 0, 0, 1, 2]
            expected_cosine_sim_scores = np.array(
                [
                    0.0000,
                    1.0000,
                    1.0000,
                    0.97464,
                    0.99819,
                    0.999618,
                ],
                dtype=np.float32,
            )
        expected_df = pd.DataFrame(
            {
                "id": expected_ids,
                "max_id": expected_max_ids,
                "cosine_sim_score": expected_cosine_sim_scores,
            }
        )
        expected_df["id"] = expected_df["id"].astype(id_col_type)
        expected_df["max_id"] = expected_df["max_id"].astype(id_col_type)
        pd.testing.assert_frame_equal(
            output_df,
            expected_df,
            check_exact=False,
            rtol=1e-5,
            atol=1e-5,
        )


@pytest.mark.gpu
class TestSemanticDedupWithoutEmbeddingCreation:
    def setup_method(self):
        self.n_clusters = 5
        self.n_samples_per_cluster = [100 * (i + 1) for i in range(self.n_clusters)]
        self.n_features = 3
        # reset all random state here to undeterministic results
        random.seed(42)
        np.random.seed(42)
        torch.manual_seed(42)
        torch.cuda.manual_seed_all(42)

        self.X, _ = make_blobs(
            n_samples=self.n_samples_per_cluster,
            centers=None,
            n_features=self.n_features,
            random_state=42,
        )
        # Convert to Dask DataFrame and then to cuDF
        df = pd.DataFrame({"id": np.arange(len(self.X)), "embeddings": self.X.tolist()})
        ddf = dd.from_pandas(df, npartitions=2)
        self.ddf = ddf.to_backend("cudf")

    def test_clustering_model(self, tmpdir):
        clustering_output_dir = os.path.join(tmpdir, "clustering_output")
        # Initialize ClusteringModel
        clustering_model = ClusteringModel(
            id_column="id",
            n_clusters=self.n_clusters,
            clustering_output_dir=clustering_output_dir,
            embedding_column="embeddings",
            random_state=42,
        )
        # TODO : remove this once we figure out why fusing is causing issues
        with dask.config.set({"optimization.fuse.active": False}):
            _ = clustering_model(DocumentDataset(self.ddf))

        # Check Directory Structure
        files = os.listdir(clustering_output_dir)
        expected_files = ["embs_by_nearest_center", "kmeans_centroids.npy"]
        for expected_file in expected_files:
            assert expected_file in files, f"The {expected_file} file should exist."

        # Check the results of embs_by_nearest_center
        for i in range(self.n_clusters):
            assert os.path.exists(
                os.path.join(
                    clustering_output_dir, "embs_by_nearest_center", f"nearest_cent={i}"
                )
            ), f"The nearest centroid file for cluster {i} should exist."
        embss_by_nearest_center = pd.read_parquet(
            os.path.join(clustering_output_dir, "embs_by_nearest_center")
        )
        # Check that embeddings are normalized
        np.testing.assert_almost_equal(
            sorted(np.stack(embss_by_nearest_center["embeddings"]).tolist()),
            sorted((self.X / np.linalg.norm(self.X, axis=1, keepdims=True)).tolist()),
        )
        num_samples_per_cluster = (
            embss_by_nearest_center["nearest_cent"].value_counts().to_dict()
        )
        np.testing.assert_allclose(
            sorted(num_samples_per_cluster.values()),
            sorted(self.n_samples_per_cluster),
            rtol=0.02,  # be within 2%
        )
        assert embss_by_nearest_center.shape[0] == len(self.X)
        assert embss_by_nearest_center.columns.tolist() == [
            "id",
            "embeddings",
            "l2_dist_to_cent",
            "cosine_dist_to_cent",
            "nearest_cent",
        ]

        # Check the results of kmeans_centroids.npy
        centroids = np.load(os.path.join(clustering_output_dir, "kmeans_centroids.npy"))
        assert centroids.shape == (self.n_clusters, self.n_features)
        # Centroids won't be normalized since they're just the mean of the embeddings
        assert not np.allclose(
            np.linalg.norm(centroids, axis=1, keepdims=True),
            np.ones_like(centroids),
        )

    def test_clustering_model_keep_all_columns(self, tmpdir):
        """Test that extra columns in the input are preserved when keep_all_columns is True."""
        clustering_output_dir = os.path.join(tmpdir, "clustering_output")
        # Create a new DataFrame with an extra column called "extra"
        df = pd.DataFrame(
            {
                "id": np.arange(len(self.X)),
                "embeddings": self.X.tolist(),
                "extra": ["foo"] * len(self.X),
            }
        )
        # Convert to Dask DataFrame and then to cuDF backend
        ddf_with_extra = dd.from_pandas(df, npartitions=2).to_backend("cudf")

        # Initialize ClusteringModel with keep_all_columns set to True.
        # We use the default sort_clusters=True to have consistent post-processing.
        clustering_model = ClusteringModel(
            id_column="id",
            n_clusters=self.n_clusters,
            clustering_output_dir=clustering_output_dir,
            embedding_column="embeddings",
            random_state=42,
            keep_all_columns=True,
        )
        # Run the clustering model on the dataset with the extra column.
        _ = clustering_model(DocumentDataset(ddf_with_extra))

        # Read the output parquet data produced after sorting clusters.
        embss_by_nearest_center = pd.read_parquet(
            os.path.join(clustering_output_dir, "embs_by_nearest_center")
        )

        # Verify that the extra column is present and its values are as expected.
        assert (
            "extra" in embss_by_nearest_center.columns
        ), "The extra column should be present when keep_all_columns is True."
        assert (
            embss_by_nearest_center["extra"].eq("foo").all()
        ), "The extra column should contain 'foo' for all rows."

    @pytest.mark.parametrize("which_to_keep", ["hard", "random", "easy"])
    @pytest.mark.parametrize("sim_metric", ["cosine", "l2"])
    def test_semantic_cluster_level_dedup(self, tmpdir, which_to_keep, sim_metric):
        clustering_output_dir = os.path.join(tmpdir, "clustering_output")
        semantic_extraction_output_dir = os.path.join(tmpdir, "extraction")
        # Initialize ClusteringModel
        clustering_model = ClusteringModel(
            id_column="id",
            n_clusters=self.n_clusters,
            clustering_output_dir=clustering_output_dir,
            embedding_column="embeddings",
            random_state=42,
        )
        with dask.config.set({"optimization.fuse.active": False}):
            _ = clustering_model(DocumentDataset(self.ddf))

        semantic_cluster_level_dedup = SemanticClusterLevelDedup(
            n_clusters=self.n_clusters,
            emb_by_clust_dir=os.path.join(
                clustering_output_dir, "embs_by_nearest_center"
            ),
            id_column="id",
            which_to_keep=which_to_keep,
            sim_metric=sim_metric,
            output_dir=semantic_extraction_output_dir,
            embedding_column="embeddings",
            batched_cosine_similarity=20,
        )

        # Call compute_semantic_match_dfs
        semantic_cluster_level_dedup.compute_semantic_match_dfs()

        output_samples_per_cluster = []
        semdedup_pruning_tables_df = []
        # Check content of semdedup_pruning_tables
        for i in range(self.n_clusters):
            cluster_i_path = os.path.join(
                semantic_extraction_output_dir,
                "semdedup_pruning_tables",
                f"cluster_{i}.parquet",
            )
            assert os.path.exists(cluster_i_path)
            df = pd.read_parquet(cluster_i_path)
            output_samples_per_cluster.append(df.shape[0])
            assert df.columns.tolist() == [
                "id",
                "max_id",
                "cosine_sim_score",
            ]
            semdedup_pruning_tables_df.append(df)

        semdedup_pruning_tables_df = pd.concat(semdedup_pruning_tables_df)
        np.testing.assert_allclose(
            sorted(output_samples_per_cluster),
            self.n_samples_per_cluster,
            rtol=0.02,  # be within 2%
        )

        # Call extract_dedup_data
        semantic_cluster_level_dedup.extract_dedup_data(eps_to_extract=0.01)
        # Check content of unique_ids
        unique_ids_path = os.path.join(
            semantic_extraction_output_dir, "unique_ids_0.01.parquet"
        )
        assert os.path.exists(unique_ids_path)
        unique_ids_df = pd.read_parquet(unique_ids_path)
        assert unique_ids_df.columns.tolist() == [
            "id",
            "cosine_dist_to_cent",
            "cluster",
        ]

        # Check content of semdedup_pruning_table with the filter matches the unique_ids
        semdedup_pruning_tables_df_filtered = semdedup_pruning_tables_df[
            semdedup_pruning_tables_df["cosine_sim_score"] >= 1 - 0.01
        ]
        assert len(semdedup_pruning_tables_df_filtered) == len(unique_ids_df)
        assert set(semdedup_pruning_tables_df_filtered["id"].to_list()) == set(
            unique_ids_df["id"].to_list()
        )

        # Check content of summary file
        summary_path = os.path.join(
            semantic_extraction_output_dir, "dedup_summary_0.01.csv"
        )
        assert os.path.exists(summary_path)
        df = pd.read_csv(summary_path)
        if which_to_keep == "hard":
            _kept, _removed = 29, 1471
        elif which_to_keep == "easy":
            _kept, _removed = 5, 1495
        else:
            _kept, _removed = 17, 1483

        pd.testing.assert_frame_equal(
            df,
            pd.DataFrame(
                {
                    "eps": [0.01],
                    "kept": [_kept],
                    "removed": [_removed],
                    "total": [len(self.X)],
                }
            ),
        )
        # Ensure that the unique_ids are also correct (this implicitly checks for semdedup_pruning_tables output)
        assert len(unique_ids_df) == _removed
        assert len(set(unique_ids_df["id"].to_list())) == _removed
